
from rest_framework import serializers

from .models import *


class EagerLoadingMixin:
    """
    Courtesy of
    http://ses4j.github.io/2015/11/23/optimizing-slow-django-rest-framework-performance/
    And see comment from Kyle Bebak for EagerLoadingMixin
    """
    @classmethod
    def setup_eager_loading(cls, queryset):
        if hasattr(cls, "_SELECT_RELATED_FIELDS"):
            queryset = queryset.select_related(*cls._SELECT_RELATED_FIELDS)
        if hasattr(cls, "_PREFETCH_RELATED_FIELDS"):
            queryset = queryset.prefetch_related(*cls._PREFETCH_RELATED_FIELDS)
        return queryset


class AddressSerializer(serializers.ModelSerializer):
    class Meta:
        model = Address
        fields = ('label', 'name', 'company', 'address1',
                  'address2', 'city', 'postal_code', 'province',
                  'country', 'phone', 'email')



import datetime

from accountifie.common.api import api_func
import accountifie.query.query_manager as QM


def balances(company_id, qstring={}):
    as_of = qstring.get('as_of', datetime.datetime.now().date().isoformat())
    return QM.QueryManager().pd_acct_balances(company_id, {as_of: as_of})[as_of].to_dict()


from django.http import HttpResponseRedirect
from django.contrib.auth.decorators import login_required

@login_required
def create_report(request):

    if 'date' in request.GET:
        dt = request.GET.get('date')
        rpt = request.GET.get('rpt')
        url = '/reporting/reports/%s?date=%s' % (rpt, dt)
    elif 'mth' in request.GET:
        mth = request.GET.get('mth')
        rpt = request.GET.get('rpt')
        yr = request.GET.get('yr')
        url = '/reporting/reports/%s?col_tag=%sM%s' % (rpt, yr, mth)
    elif 'qtr' in request.GET:
        qtr = request.GET.get('qtr')
        rpt = request.GET.get('rpt')
        yr = request.GET.get('yr')
        url = '/reporting/reports/%s?col_tag=%sQ%s' % (rpt, yr, qtr)    
    else:
        rpt = request.GET.get('rpt')
        yr = request.GET.get('yr')
        period = request.GET.get('period')
        url = '/reporting/reports/%s?col_tag=%s%s' % (rpt, yr, period)    

    return HttpResponseRedirect(url)



from django.conf import settings
from django.conf.urls import include, url, handler404, handler500
from django.contrib.staticfiles.urls import staticfiles_urlpatterns

from . import views


urlpatterns = [
    url(r'snapshot/glsnapshots/$', views.glsnapshots, name='glsnapshots'),
    url(r'snapshot/glsnapshots/balances/(?P<snap_id>[()_a-zA-Z0-9]+)/$', views.glsnapshots_balances),   
]



"""
Adapted with permission from ReportLab's DocEngine framework

An easy way to provide useful stats.  Put a module like this in your app, with whatever
queries or callables you want in it.  They should return list-of-lists structures.


"""

SQL_QUERIES = [
    [
    'all_users',
    'lists all user names and email addresses',
    'SELECT id, username, email, is_staff, is_superuser FROM auth_user'
    ],

]


# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-11-27 01:10
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('snapshot', '0002_auto_20160414_1820'),
    ]

    operations = [
        migrations.AlterField(
            model_name='glsnapshot',
            name='snapped_at',
            field=models.DateTimeField(help_text='Local NY time'),
        ),
    ]



import datetime
from decimal import Decimal
import pandas as pd
import json
import csv

from django.contrib.auth.decorators import login_required
from django.shortcuts import render, get_object_or_404
from django.template import RequestContext
from django.views.generic.edit import CreateView
from django.core.urlresolvers import reverse, reverse_lazy
from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponse

from accountifie.cal.models import Year

from . import models
from accountifie.common.api import api_func
from accountifie.query.query_manager_strategy_factory import QueryManagerStrategyFactory
from accountifie.query.query_manager import QueryManager
from accountifie.reporting.rptutils import report_prep
import accountifie.toolkit.utils as utils
from accountifie.common.table import get_table

import logging
logger = logging.getLogger('default')

@login_required
def glsnapshots(request):
    
    context = {}
    context['title'] = 'Snapshots'
    context['content'] = get_table('snapshots')()
    return render(request, 'snapshot/base_snapshot.html', context)



def glsnapshots_balances(request, snap_id):
    snapshot = models.GLSnapshot.objects.get(id=snap_id)
    snapshot_time = snapshot.snapped_at.strftime('%Y-%m-%dT%H:%M:%S.0000ZZ')

    strategy = QueryManagerStrategyFactory().get('snapshot')
    strategy.set_cache(snapshot_time)

    report, is_report, format = report_prep(request, 'RecBalances')
    if not is_report:
        return report

    report.date = snapshot.closing_date
    report.qm_strategy = strategy
    report.snapshot = snapshot
    report.columns = {'snapshot': 'snapshot', 'current': 'current', 'diff': 'diff'}

    report_data = report.calcs()
    
    if format == 'json':
        return HttpResponse(json.dumps(report_data, cls=DjangoJSONEncoder), content_type="application/json")
    elif format == 'csv':
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename="report.csv"'

        writer = csv.writer(response)
        writer.writerow([''] + report.column_order)
        map_values = lambda x: '' if x=='' else str(x['text']).replace('-','0')
        for row in report_data:
            writer.writerow([row['label']] + [map_values(row[col]) for col in report.column_order])
        return response
    elif format == 'html':
        # not bootstrap ... the more custom style for more complex reports
        context = report.html_report_context()
        context['query_string']=request.META['QUERY_STRING']
        context['rows'] = []
        for rec in report_data:
            context['rows'] += report.get_row(rec)

        return render(request, 'report.html', context)
    else:
        msg = "Sorry. This format is not recognised : %s" % format
        return render(request, '404.html', {'message': msg}), False



import datetime

from django.template import Library

import accountifie.toolkit.utils as utils

register = Library()


@register.simple_tag
def start_of_year():
    return utils.start_of_year(datetime.datetime.now().year).isoformat()




from accountifie.reporting.models import MetricEntry, ReportDef
from rest_framework import serializers


class MetricEntrySerializer(serializers.ModelSerializer):
    metric = serializers.StringRelatedField()

    class Meta:
        model = MetricEntry
        fields = ('metric', 'label', 'date', 'balance')


class ReportDefSerializer(serializers.ModelSerializer):
    class Meta:
        model = ReportDef
        fields = ('name', 'description', 'path', 'calc_type')



import datetime
import pandas as pd

from django.conf import settings

from accountifie.common.api import api_func
import accountifie.query.query_manager as QM
from accountifie.toolkit.utils import start_of_year, start_of_prev_year, daterange
import accountifie.reporting.rptutils as rptutils

import logging

logger = logging.getLogger('default')


def transaction(trans_id, qstring={}):
    # no way right now to know whether which company it is in
    companies = [cmpy['id'] for cmpy in api_func('gl', 'company') if cmpy['cmpy_type']=='ALO']
    
    for cmpny in companies:
        info = QM.QueryManager().transaction_info(cmpny, trans_id)

        if len(info)>0:
            return info
    
    return None


def _is_path(path):
    return (path.split('.')[0] in ['assets', 'liabilities', 'equity'])

def _account_history(acct_id, company_id, from_date, to_date, cp):
    return QM.QueryManager().pd_history(company_id, 'account', acct_id, from_date=from_date, to_date=to_date, cp=cp)


def _cp_history(cp_id, company_id , from_date, to_date):
    return QM.QueryManager().pd_history(company_id, 'account', '3000', from_date=from_date, to_date=to_date, cp=cp_id)
        
    
def _path_history(path, company_id , from_date, to_date, excl, incl):
    return QM.QueryManager().pd_history(company_id, 'path', path, from_date=from_date, to_date=to_date, excl_contra=excl, incl=incl)

def _cutoff(start_cutoff, hist):
    if hist.empty:
        return pd.DataFrame()
    
    used_history = hist[hist['date'] >= start_cutoff]
    unused_history = hist[hist['date'] < start_cutoff][-1:]
    
    unused_history['date'] = start_cutoff
    unused_history['id'] = 'Start'
    for col in [x for x in list(unused_history.columns) if x not in ['id','date','balance']]:
        unused_history[col] = ''
    unused_history['amount'] = 0.0
    unused_history['comment'] = 'Starting Balance'
    
    return pd.concat([unused_history, used_history])
    
def history(id, qstring={}):

    end_date = rptutils.date_from_shortcut(qstring.get('to', datetime.datetime.now().date()))
    start_cutoff = qstring.get('from', None)

    if start_cutoff:
        start_cutoff = rptutils.date_from_shortcut(start_cutoff)
    else:
        start_cutoff = start_of_prev_year(end_date.year)
    start_date = settings.DATE_EARLY
    
    cp = qstring.get('cp', None)
    excl = qstring.get('excl', None)
    incl = qstring.get('incl', None)

    company_id = qstring.get('company_id', api_func('environment', 'variable', 'DEFAULT_COMPANY_ID'))
    
    if api_func('gl', 'account', str(id)) is not None:
        hist = _account_history(id, company_id, start_date, end_date, cp)
        return _cutoff(start_cutoff, hist).to_dict(orient='records')
    elif api_func('gl', 'counterparty', id) is not None:
        hist = _cp_history(id, company_id, start_date, end_date)
        return _cutoff(start_cutoff, hist).to_dict(orient='records')
    elif _is_path(id):
        hist = _path_history(id, company_id, start_date, end_date, excl, incl)
        return _cutoff(start_cutoff, hist).to_dict(orient='records')
    else:
        return "ID %s not recognized as an account or counterparty" %id


def balance_history(id, qstring={}):
    end_date = rptutils.date_from_shortcut(qstring.get('to_date', datetime.datetime.now().date()))

    start_cutoff = qstring.get('from_date', None)
    if start_cutoff:
        start_cutoff = rptutils.date_from_shortcut(start_cutoff)
    else:
        start_cutoff = start_of_year(end_date.year)
    start_date = settings.DATE_EARLY

    cp = qstring.get('cp', None)
    excl = qstring.get('excl', None)
    incl = qstring.get('incl', None)

    if not start_cutoff:
        start_cutoff = start_of_year(end_date.year)

    company_id = qstring.get('company_id', api_func('environment', 'variable', 'DEFAULT_COMPANY_ID'))
    
    if api_func('gl', 'account', str(id)) is not None:
        hist = _account_history(str(id), company_id, start_date, end_date, cp)
        hist = _cutoff(start_cutoff, hist)
    elif api_func('gl', 'counterparty', id) is not None:
        hist = _cp_history(id, company_id, start_date, end_date)
        hist = _cutoff(start_cutoff, hist)
    elif _is_path(id):
        hist = _path_history(id, company_id, start_date, end_date, excl, incl)
        hist = _cutoff(start_cutoff, hist)
    else:
        return "ID %s not recognized as an account or counterparty" %id

    hist['idx'] = hist.index
    date_indices = hist[['date','idx']].groupby('date').max()
    balances = hist.loc[date_indices['idx']][['date','balance']]
    balances.index = balances['date']
    del balances['date']

    dates = daterange(balances.index[0], balances.index[-1], bus_days_only=True)
    filled_balances = pd.DataFrame(index=dates)
    filled_balances['balance'] = balances['balance']
    filled_balances.fillna(method='ffill', inplace=True)
    filled_balances.index = filled_balances.index.map(lambda x: x.isoformat())

    return filled_balances['balance'].to_dict()



import locale

from django.template import Library

from accountifie.gl.models import Company

register = Library()

@register.filter
def currency(value, arg = '', symbol = True):
    '''
    Currency formatting template filter.

    Takes a number -- integer, float, decimal -- and formats it according to
    the locale specified as the template tag argument (arg). Examples:

      * {{ value|currency }}
      * {{ value|currency:"en_US" }}
      * {{ value|currency:"pt_BR" }}
      * {{ value|currency:"pt_BR.UTF8" }}

    If the argument is omitted, the default system locale will be used.

    The third parameter, symbol, controls whether the currency symbol will be
    printed or not. Defaults to true.

    As advised by the Django documentation, this template won't raise
    exceptions caused by wrong types or invalid locale arguments. It will
    return an empty string instead.

    Be aware that currency formatting is not possible using the 'C' locale.
    This function will fall back to 'en_US.UTF8' in this case.
    '''

    saved = locale.getlocale()
    given = arg and ('.' in arg and str(arg) or (arg, 'UTF8'))

    if saved == (None, None) and given == '':
        given = 'en_US.UTF8'

    try:
        locale.setlocale(locale.LC_ALL, given)

        return locale.currency(value, symbol, True)

    except (TypeError, locale.Error):
        return ''

    finally:
        locale.setlocale(locale.LC_ALL, saved)

@register.filter
def company_color(cid):
    try:
        company = Company.objects.get(pk=cid)
        return company.color_code or '#000'
    except Company.DoesNotExist:
        return '#000'



import pkgutil
import inspect
import datetime
from dateutil.parser import parse

from django.conf import settings
from django.shortcuts import render
from django.template import RequestContext

import accountifie.toolkit.utils as utils
from accountifie.reporting.models import ReportDef
from . import column_funcs as colfuncs
import accountifie.toolkit.utils.datefuncs as datefuncs
from . import shortcuts


def qs_parse(qs):
    matches = []
    if 'date' in qs:
        matches.append('date')
    
    if 'from' in qs or 'to' in qs:
        matches.append('date_range')

    if 'period' in qs and 'by' in qs:
        matches.append('period')

    if 'col_tag' in qs:
        matches.append('shortcut')

    return matches


def config_fromdate(calc_type, rpt_desc, dt):
    dt = shortcuts.date_from_shortcut(dt)

    if calc_type == 'as_of':
        as_of_col = {dt.strftime('%d-%b-%y'): dt.isoformat()}
    else:
        as_of_col = {str(dt.year): str(dt.year)}
    
    config = {}
    config['columns'] = as_of_col
    config['column_order'] = list(as_of_col.keys())
    config['title'] = '%s, %s' % (rpt_desc, dt.strftime('%d-%b-%y'))
    config['date'] = dt
    return config


def config_fromperiod(calc_type, rpt_desc, config):
    period_id = colfuncs.get_period_id(config)

    columns, column_titles = colfuncs.gen_periods(calc_type, config)
    title = '%s %s' % (period_id, rpt_desc)
    return {'title': title, 'columns': dict(list(zip(column_titles, columns))), 'column_order': column_titles}


def config_fromdaterange(calc_type, rpt_desc, config):
    columns, column_titles = colfuncs.daterange_periods(calc_type, config)
    title = 'From %s to %s. %s' % (config['from'], config['to'], rpt_desc)
    return {'title': title, 'columns': dict(list(zip(column_titles, columns))), 'column_order': column_titles}


def get_report(rpt_id, company_id, version=None):
    try:
        report = ReportDef.objects.get(name=rpt_id)
        rpt_path = report.path
        rpt_name = report.name
    except:
        return None
    
    loader = pkgutil.get_loader(rpt_path)
    module = loader.load_module(rpt_path)
    
    rpt = None
    for name, obj in inspect.getmembers(module):
        if name == rpt_name:
            rpt = obj(company_id)
            break
    return rpt


def history_prep(request):
    config = dict((k, request.GET.get(k)) for k in list(request.GET.keys()))
    
    if config.get('company_id') is None:
        config['company_id'] = utils.get_company(request)

    qs_matches = qs_parse(config)

    if len(qs_matches) == 0:
        config['config_type'] = 'date_range'
        config['from'] = settings.DATE_EARLY
        config['to'] = settings.DATE_LATE
        
    elif len(qs_matches) > 1:
        raise ValueError('Unexpected query string: %s' % repr(config))
    else:
        config['config_type'] = qs_matches[0]

    if config['config_type'] == 'shortcut':
        config.pop('config_type')
        config.update(shortcuts.parse_shortcut(config['col_tag']))

    if config['config_type'] == 'date_range':
        if config.get('from'): 
            config['from'] = shortcuts.date_from_shortcut(config.get('from'))
        if config.get('to'):
            config['to'] = shortcuts.date_from_shortcut(config.get('to'))
    elif config['config_type'] == 'period':
        period_id = colfuncs.get_period_id(config)
        config['from'] = datefuncs.start_of_period(period_id)
        config['to'] = datefuncs.end_of_period(period_id)
        
    return config


def report_prep(request, id):
    
    qs = dict((k, request.GET.get(k)) for k in list(request.GET.keys()))
    if qs.get('company_id') is None:
        qs['company_id'] = utils.get_company(request)

    report = get_report(id, qs['company_id'])
    gl_strategy = request.GET.get('gl_strategy', None)

    if report is None:
        msg = "Report %s does not exist" % id
        return render(request, 'rpt_doesnt_exist.html', {'message': msg}), False

    if qs['company_id'] not in report.works_for:
        msg = "This ain't it. Report not available for %s" % report.company_name
        return render(request, 'rpt_doesnt_exist.html', {'message': msg}), False

    report.configure(qs)
    report.format = qs.get('format', 'html')
    report.set_gl_strategy(gl_strategy)

    return report, True





"""
Strategy that delegates transaction querying to accountifie-svc.

The new kid on the block, capable of going from 0 to 100km/h in under 50ms
though with the chance of a few early speed wobbles.

See query_manager_strategy.py for interface docs.
Use query_manager_strategy_factory.py to get an instance of this class.
"""

import pandas as pd
import itertools
import json
import urllib.request, urllib.parse, urllib.error
import urllib.request, urllib.error, urllib.parse
from decimal import Decimal
from collections import defaultdict

from dateutil.parser import parse
from django.conf import settings
from .query_manager_strategy import QueryManagerStrategy
from accountifie.common.api import api_func
import logging

import time
DZERO = Decimal('0')
logger = logging.getLogger('default')


class QueryManagerRemoteStrategy(QueryManagerStrategy):
    def cp_balances_for_dates(self, company_id, account_ids, dates):        
        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            company_list = api_func('gl', 'company_list', company_id)
            balances = [self.cp_balances_for_dates(cmpny, account_ids, dates) for cmpny in company_list]
            return self.__merge_cp_balances_for_dates_results(balances)
        
        client = accountifieSvcClient(company_id)
        result = defaultdict(dict)

        for dt in dates:
            start = dates[dt]['start']
            end = dates[dt]['end']
            cp_balances = client.cp_balances(accounts=account_ids,
                                             from_date=start,
                                             to_date=end)

            for acct_bals in cp_balances:
                result[acct_bals['id']][dt] = {'openingBalance': acct_bals['openingBalance'],
                                               'closingBalance': acct_bals['closingBalance']}

        return result

    def account_balances_for_dates(self, company_id, account_ids, dates, with_counterparties, excl_interco, excl_contra, with_tags, excl_tags):
        
        interco_exempt_accounts = api_func('gl', 'externalaccounts')

        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            company_list = api_func('gl', 'company_list', company_id)
            balances = [self.account_balances_for_dates(cmpny, account_ids, dates, with_counterparties, True, excl_contra, with_tags, excl_tags) for cmpny in company_list]
            return self.__merge_account_balances_for_dates_results(balances)

        
        client = accountifieSvcClient(company_id)
        date_indexed_account_balances = {}

        interco_counterparties = self.get_interco_counterparties_for(company_id) if excl_interco else []
        accounts_with_interco_exclusion =    [id for id in account_ids if id not in interco_exempt_accounts] if excl_interco else []
        accounts_without_interco_exclusion = [id for id in account_ids if id in interco_exempt_accounts]     if excl_interco else account_ids

        for dt in dates:
            start = dates[dt]['start']
            end = dates[dt]['end']

            interco_excluded_balances = client.balances(
                accounts=accounts_with_interco_exclusion,
                from_date=start,
                to_date=end,
                with_counterparties=with_counterparties,
                excluding_counterparties=interco_counterparties,
                excluding_contra_accounts=excl_contra,
                with_tags=with_tags,
                excluding_tags=excl_tags
            ) if len(accounts_with_interco_exclusion) > 0 else []

            interco_included_balances = client.balances(
                accounts=accounts_without_interco_exclusion,
                from_date=start,
                to_date=end,
                with_counterparties=with_counterparties,
                excluding_contra_accounts=excl_contra,
                with_tags=with_tags,
                excluding_tags=excl_tags
            ) if len(accounts_without_interco_exclusion) > 0 else []

            balances = interco_excluded_balances + interco_included_balances
            date_indexed_account_balances[dt] = dict((balance['id'], float(balance['shift'])) for balance in balances)        
        return date_indexed_account_balances

    def transactions(self, company_id, account_ids, from_date, to_date, chunk_frequency, with_counterparties, excl_interco, excl_contra):
        interco_exempt_accounts = api_func('gl', 'externalaccounts')

        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            company_list = api_func('gl', 'company_list', company_id)
            balances = [self.transactions(cmpny, account_ids, from_date, to_date, chunk_frequency, with_counterparties, True, excl_contra) for cmpny in company_list]
            return self.__merge_transactions_results(balances)

        client = accountifieSvcClient(company_id)

        interco_counterparties = self.get_interco_counterparties_for(company_id) if excl_interco else []
        accounts_with_interco_exclusion =    [id for id in account_ids if id not in interco_exempt_accounts] if excl_interco else []
        accounts_without_interco_exclusion = [id for id in account_ids if id in interco_exempt_accounts]     if excl_interco else account_ids

        interco_excluded_transactions = client.transactions(
            accounts=accounts_with_interco_exclusion,
            from_date=from_date,
            to_date=to_date,
            chunk_frequency=chunk_frequency,
            excluding_counterparties=interco_counterparties,
            with_counterparties=with_counterparties,
            excluding_contra_accounts=excl_contra
        ) if len(accounts_with_interco_exclusion) > 0 else {}

        interco_included_transactions = client.transactions(
            accounts=accounts_without_interco_exclusion,
            from_date=from_date,
            to_date=to_date,
            chunk_frequency=chunk_frequency,
            with_counterparties=with_counterparties,
            excluding_contra_accounts=excl_contra
        ) if len(accounts_without_interco_exclusion) > 0 else {}

        # merge the two dictionaries
        transactions = interco_excluded_transactions.copy()
        transactions.update(interco_included_transactions)

        make_contras = lambda x: 'more than two' if len(x)>2 else '.'.join(x)

        formatted_transactions = []
        for account in transactions:
            for transaction in transactions[account]:
                if transaction['amount'] != '0.00':
                    formatted_transactions.append({
                        'date': parse(transaction['dateEnd'] if ('dateEnd' in transaction and transaction['dateEnd']) else transaction['date']).date(),
                        'id': transaction['id'],
                        'comment': transaction['comment'],
                        'account_id': account,
                        'contra_accts': make_contras(transaction['contraAccounts']),
                        'counterparty': transaction['counterparty'],
                        'amount': float(transaction['amount'])
                    })

        return formatted_transactions

    def __merge_cp_balances_for_dates_results(self, result_list):
        merged_results = {}

        for result in result_list:
            for acct in result:
                merged_results[acct] = merged_results[acct] if acct in merged_results else {}
                for date in result[acct]:
                    if date not in merged_results[acct]:
                        merged_results[acct][date] = {'closingBalance': [], 'openingBalance': []}
                    existing_amount = merged_results[acct][date]
                    for col in ['closingBalance', 'openingBalance']:
                        all_rows = existing_amount[col] + result[acct][date][col]
                        merged_results[acct][date][col] = []
                        for k,v in itertools.groupby(sorted(all_rows, key=lambda x: x['cp']), key=lambda x: x['cp']):
                            merged_results[acct][date][col].append({'cp': k, 'total': str(sum(Decimal(r['total']) for r in v))})
        return merged_results

    def __merge_account_balances_for_dates_results(self, result_list):
        merged_results = {}

        for result in result_list:
            for date in result:
                merged_results[date] = merged_results[date] if date in merged_results else {}
                for account in result[date]:
                    existing_amount = merged_results[date][account] if account in merged_results[date] else float(0)
                    merged_results[date][account] = existing_amount + result[date][account]

        return merged_results

    def __merge_transactions_results(self, result_list):
        return [transaction for result in result_list for transaction in result]

    def get_transaction(self, company_id, transaction_id):
        client = accountifieSvcClient(company_id)
        return client.get_transaction(transaction_id)

    def create_gl_transactions(self, d2, lines, trans_id, bmo_id):
        if sum([Decimal(line['amount']) for line in lines]) == DZERO:
            self.upsert_transaction({
                'id': trans_id,
                'bmo_id': bmo_id,
                'date': str(d2['date']),
                'date_end': str(d2.get('date_end', None) or d2['date']),
                'comment': d2['comment'],
                'company': d2['company'],
                'type': None,
                'lines': [dict(account=line['account'],
                                amount=str(Decimal(line['amount'])),
                                counterparty = line['counterparty'],
                                tags = line['tags'] ) for line in lines]
            })
        else:
            logger.info('Unbalanced entry for bmo ID: %s' % bmo_id)


    def upsert_transaction(self, transaction):
        client = accountifieSvcClient(transaction['company'])
        client.upsert_transaction(transaction)

    def delete_transaction(self, company_id, transaction_id):
        client = accountifieSvcClient(company_id)
        client.delete_transaction(transaction_id)

    def delete_bmo_transactions(self, company_id, bmo_id):
        client = accountifieSvcClient(company_id)
        client.delete_bmo_transactions(bmo_id)

    def take_snapshot(self, company_id, snapshot_time=None):
        accountifieSvcClient(company_id).take_snapshot(snapshot_time)

    def erase(self, company_id):
        accountifieSvcClient(company_id).erase()

    def set_fast_inserts(self, company_id, value):
        if value:
            accountifieSvcClient(company_id).disable_balance_cache()
        else:
            accountifieSvcClient(company_id).enable_balance_cache()


    @staticmethod
    def __inter_co(row):
        ext_accts = api_func('gl', 'externalaccounts')
        companies = [cmpy['id'] for cmpy in api_func('gl', 'company') if cmpy['id']!=company]
        if row['account_id'] in ext_accts:
            return False
        if row['counterparty'] in companies:
            return True
        else:
            return False

    @staticmethod
    def get_interco_counterparties_for(company):
        output = [cmpy['id'] for cmpy in api_func('gl', 'company') if cmpy['id']!=company and cmpy['cmpy_type']=='ALO']
        output += [x.lower() for x in output]
        return output

class accountifieSvcClient(object):
    def __init__(self, company_id='*'):
        self.company_id = company_id
        self.url_base = "%s/gl/%s" % (self.__accountifie_svc_url(), company_id)

    def __post(self, path, params=None):
        params = json.dumps(params) if params is not None else '{}'
        url = '%s%s' % (self.url_base, path)
        request = urllib.request.Request(url, data=params, headers={'Content-Type': 'application/json'})
        try:
            response = urllib.request.urlopen(request)
            json_result = json.load(response)
            return json_result
        except:
            logger.error('Accountifie svc post failed on %s with params %s', url, str(params))
            return None

    def __get(self, path, params=None):
        if params:
            params = urllib.parse.urlencode(params)
            url = '%s%s?%s' % (self.url_base, path, params)
        else:
            url = '%s%s' % (self.url_base, path)

        start_time = time.time()
        request = urllib.request.Request(url)
        response = urllib.request.urlopen(request)
        json_result = json.load(response)

        resp_time = time.time() - start_time
        if resp_time > 2:
            logger.info('Accountifie-svc. Long response time. %.2f seconds. URL=%s' % (resp_time, url))

        return json_result

    def balances(self, accounts, from_date=None, to_date=None, with_counterparties=None, excluding_counterparties=None, excluding_contra_accounts=None, with_tags=None, excluding_tags=None):
        from_date = None if from_date == '2000-01-01' else from_date
        account_balances = self.__get('/balances', {
            'accounts': ','.join(accounts),
            'from': from_date,
            'to': to_date,
            'withCounterparties': ','.join(with_counterparties or []),
            'excludingCounterparties': ','.join(excluding_counterparties or []),
            'excludingContraAccounts': ','.join(excluding_contra_accounts or []),
            'withTags': ','.join(with_tags or []),
            'excludingTags': ','.join(excluding_tags or []),
        })

        return account_balances

    def cp_balances(self, accounts, from_date=None, to_date=None):
        from_date = None if from_date == '2000-01-01' else from_date
        cp_balances = self.__get('/cpBalances', {
            'accounts': ','.join(accounts),
            'from': from_date,
            'to': to_date
        })

        return cp_balances

    def transactions(self, accounts, from_date=None, to_date=None, chunk_frequency=None, with_counterparties=None, excluding_counterparties=None, excluding_contra_accounts=None):
        from_date = None if from_date == '2000-01-01' else from_date
        transaction_history = self.__get('/transactions', {
            'accounts': ','.join(accounts),
            'from': from_date,
            'to': to_date,
            'chunkFrequency': chunk_frequency,
            'withCounterparties': ','.join(with_counterparties or []),
            'excludingCounterparties': ','.join(excluding_counterparties or []),
            'excludingContraAccounts': ','.join(excluding_contra_accounts or [])
        })
        return transaction_history


    def get_transaction(self, transaction_id):
        return self.__get('/transaction/%s' % transaction_id)


    def upsert_transaction(self, transaction):
        self.__post('/transaction/%s/upsert' % transaction['id'], {
            'id': transaction['id'],
            'bmoId': transaction['bmo_id'],
            'date': transaction['date'],
            'dateEnd': transaction['date_end'],
            'comment': transaction['comment'],
            'type': transaction['type'],
            'source_object': transaction['source_object'],
            'lines': [{'accountId': line['account'],
                       'amount': line['amount'],
                       'counterpartyId': line['counterparty'],
                       'tags': line['tags']
                       }
                      for line in transaction['lines']
                      ]
        })

    def delete_transaction(self, transaction_id):
        self.__post('/transaction/%s/delete' % transaction_id)

    def delete_bmo_transactions(self, bmo_id):
        self.__post('/bmo-transactions/%s/delete' % bmo_id)

    def erase(self):
        self.__post('/erase')

    def enable_balance_cache(self):
        self.__post('/enable-balance-cache')

    def disable_balance_cache(self):
        self.__post('/disable-balance-cache')

    def take_snapshot(self, snapshot_time):
        logger.info('taking snapshot for %s with snapshot_time=%s' % (self.company_id, snapshot_time))
        if snapshot_time:
            self.__post('/snapshot/create?date=%s' % snapshot_time)
        else:
            self.__post('/snapshot/create')

    @staticmethod
    def __accountifie_svc_url():
        return api_func('environment', 'variable', 'ACCOUNTIFIE_SVC_URL')



from django.contrib.auth.decorators import login_required
from django.shortcuts import render

from accountifie.common.uploaders.upload_tools import order_upload
from accountifie.toolkit.forms import LabelledFileForm
import accountifie.reporting.importers

@login_required
def upload_file(request, file_type, check=False):
	if request.method == 'POST':
		if file_type == 'metrics':
			processor = accountifie.reporting.importers.process_metrics
			return order_upload(request,
								processor,
								label=True,
								redirect_url='/projections/parameters/')
		else:
			raise ValueError("Unexpected file type; know about metrics")
	else:
		form = LabelledFileForm()
		context = {'form': form, 'file_type': file_type}
		return render(request, 'common/upload_csv.html', context)



from django.contrib import admin

from .models import Metric, MetricEntry, ReportDef

class MetricAdmin(admin.ModelAdmin):
    list_display = ('name',)

class MetricEntryAdmin(admin.ModelAdmin):
    list_display = ('metric', 'date', 'balance', 'label', 'as_of')
    list_filter = ('metric', 'label')

class ReportDefAdmin(admin.ModelAdmin):
    list_display = ('name', 'description', 'path', 'calc_type')


admin.site.register(Metric, MetricAdmin)
admin.site.register(MetricEntry, MetricEntryAdmin)
admin.site.register(ReportDef, ReportDefAdmin)



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='GLSnapshot',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('snapped_at', models.DateTimeField()),
                ('short_desc', models.CharField(max_length=100)),
                ('comment', models.TextField()),
                ('closing_date', models.DateField()),
            ],
        ),
    ]



from multipledispatch import dispatch

from accountifie.gl.models import Company


@dispatch(str, dict)
def company(company_id, qstring):
    company = Company.objects.get(id=company_id)
    flds = ['cmpy_type', 'color_code', 'name', 'id']
    data = dict((str(k),str(v)) for k,v in company.__dict__.items() if k in flds)

    if company.cmpy_type == 'CON':
        data['subs'] = [sub.id for sub in company.subs.all()]

    return data

@dispatch(dict)
def company(qstring):
    company_list = list(Company.objects.all())
    flds = ['cmpy_type', 'color_code', 'name', 'id']

    data = []

    for company in company_list:
        company_data = dict((str(k),str(v)) for k,v in company.__dict__.items() if k in flds)

        if company.cmpy_type == 'CON':
            company_data['subs'] = [sub.id for sub in company.subs.all()]

        data.append(company_data)

    return data

"""
def companies(qstring={}):
    company_list = list(Company.objects.all())
    flds = ['cmpy_type', 'color_code', 'name', 'id']

    data = []

    for company in company_list:
        company_data = dict((str(k),str(v)) for k,v in company.__dict__.iteritems() if k in flds)

        if company.cmpy_type == 'CON':
            company_data['subs'] = [sub.id for sub in company.subs.all()]

        data.append(company_data)

    return data
"""


def company_list(company_id, qstring={}):
    company = Company.objects.get(id=company_id)
    if company.cmpy_type == 'CON':
        return [sub.id for sub in company.subs.all()]
    else:
        return [company_id]


"""
Strategy that delegates transaction querying to a snapshot of accountifie-svc.

It is closely modelled on the remote Strategy
See query_manager_strategy.py for interface docs.
Use query_manager_strategy_factory.py to get an instance of this class.
"""

import time
import json
import urllib.request, urllib.parse, urllib.error
import urllib.request, urllib.error, urllib.parse
from decimal import Decimal

from dateutil.parser import parse
from django.conf import settings
from .query_manager_strategy import QueryManagerStrategy
from accountifie.common.api import api_func
import logging

DZERO = Decimal('0')
logger = logging.getLogger('default')


class QueryManagerSnapshotStrategy(QueryManagerStrategy):

    def set_cache(self, snapshot_time):
        self.snapshot_time = snapshot_time


    def account_balances_for_dates(self, company_id, account_ids, dates, with_counterparties, excl_interco, excl_contra, with_tags, excl_tags):
        interco_exempt_accounts = api_func('gl', 'externalaccounts')

        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            company_list = api_func('gl', 'company_list', company_id)
            balances = [self.account_balances_for_dates(cmpny, account_ids, dates, with_counterparties, True, excl_contra, with_tags, excl_tags) for cmpny in company_list]
            return self.__merge_account_balances_for_dates_results(balances)

        client = accountifieSvcSnapshotClient(self.snapshot_time, company_id=company_id)
        date_indexed_account_balances = {}

        interco_counterparties = self.get_interco_counterparties_for(company_id) if excl_interco else []
        accounts_with_interco_exclusion =    [id for id in account_ids if id not in interco_exempt_accounts] if excl_interco else []
        accounts_without_interco_exclusion = [id for id in account_ids if id in interco_exempt_accounts]     if excl_interco else account_ids

        for dt in dates:
            start = dates[dt]['start']
            end = dates[dt]['end']

            interco_excluded_balances = client.balances(
                accounts=accounts_with_interco_exclusion,
                from_date=start,
                to_date=end,
                with_counterparties=with_counterparties,
                excluding_counterparties=interco_counterparties,
                excluding_contra_accounts=excl_contra,
                with_tags=with_tags,
                excluding_tags=excl_tags
            ) if len(accounts_with_interco_exclusion) > 0 else []

            interco_included_balances = client.balances(
                accounts=accounts_without_interco_exclusion,
                from_date=start,
                to_date=end,
                with_counterparties=with_counterparties,
                excluding_contra_accounts=excl_contra,
                with_tags=with_tags,
                excluding_tags=excl_tags
            ) if len(accounts_without_interco_exclusion) > 0 else []

            balances = interco_excluded_balances + interco_included_balances
            date_indexed_account_balances[dt] = dict((balance['id'], float(balance['shift'])) for balance in balances)        
        return date_indexed_account_balances

    """
    def transactions(self, company_id, account_ids, from_date, to_date, chunk_frequency, with_counterparties, excl_interco, excl_contra):
        interco_exempt_accounts = api_func('gl', 'externalaccounts')

        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            company_list = api_func('gl', 'company_list', company_id)
            balances = [self.transactions(cmpny, account_ids, from_date, to_date, chunk_frequency, with_counterparties, True, excl_contra) for cmpny in company_list]
            return self.__merge_transactions_results(balances)

        client = accountifieSvcSnapshotClient(company_id)

        interco_counterparties = self.get_interco_counterparties_for(company_id) if excl_interco else []
        accounts_with_interco_exclusion =    [id for id in account_ids if id not in interco_exempt_accounts] if excl_interco else []
        accounts_without_interco_exclusion = [id for id in account_ids if id in interco_exempt_accounts]     if excl_interco else account_ids

        interco_excluded_transactions = client.transactions(
            accounts=accounts_with_interco_exclusion,
            from_date=from_date,
            to_date=to_date,
            chunk_frequency=chunk_frequency,
            excluding_counterparties=interco_counterparties,
            with_counterparties=with_counterparties,
            excluding_contra_accounts=excl_contra
        ) if len(accounts_with_interco_exclusion) > 0 else {}

        interco_included_transactions = client.transactions(
            accounts=accounts_without_interco_exclusion,
            from_date=from_date,
            to_date=to_date,
            chunk_frequency=chunk_frequency,
            with_counterparties=with_counterparties,
            excluding_contra_accounts=excl_contra
        ) if len(accounts_without_interco_exclusion) > 0 else {}

        # merge the two dictionaries
        transactions = interco_excluded_transactions.copy()
        transactions.update(interco_included_transactions)

        make_contras = lambda x: 'more than two' if len(x)>2 else '.'.join(x)

        formatted_transactions = []
        for account in transactions:
            for transaction in transactions[account]:
                if transaction['amount'] != '0.00':
                    formatted_transactions.append({
                        'date': parse(transaction['dateEnd'] if ('dateEnd' in transaction and transaction['dateEnd']) else transaction['date']).date(),
                        'id': transaction['id'],
                        'comment': transaction['comment'],
                        'account_id': account,
                        'contra_accts': make_contras(transaction['contraAccounts']),
                        'counterparty': transaction['counterparty'],
                        'amount': float(transaction['amount'])
                    })

        return formatted_transactions
    """

    def __merge_account_balances_for_dates_results(self, result_list):
        merged_results = {}

        for result in result_list:
            for date in result:
                merged_results[date] = merged_results[date] if date in merged_results else {}
                for account in result[date]:
                    existing_amount = merged_results[date][account] if account in merged_results[date] else float(0)
                    merged_results[date][account] = existing_amount + result[date][account]

        return merged_results

    """
    def __merge_transactions_results(self, result_list):
        return [transaction for result in result_list for transaction in result]

    def get_transaction(self, company_id, transaction_id):
        client = accountifieSvcSnapshotClient(company_id)
        return client.get_transaction(transaction_id)
    """
    


    @staticmethod
    def __inter_co(row):
        ext_accts = api_func('gl', 'externalaccounts')
        companies = [cmpy['id'] for cmpy in api_func('gl', 'company') if cmpy['id']!=company]
        if row['account_id'] in ext_accts:
            return False
        if row['counterparty'] in companies:
            return True
        else:
            return False

    @staticmethod
    def get_interco_counterparties_for(company):
        output = [cmpy['id'] for cmpy in api_func('gl', 'company') if cmpy['id']!=company and cmpy['cmpy_type']=='ALO']
        output += [x.lower() for x in output]
        return output

class accountifieSvcSnapshotClient(object):
    def __init__(self, snapshot_time, company_id='*'):
        self.company_id = company_id
        self.url_base = "%s/gl/%s/snapshot" % (self.__accountifie_svc_url(), company_id)
        self.snapshot_time = snapshot_time

    def __get(self, path, params=None):
        if params:
            params.update({'snapshotDate': self.snapshot_time})
        else:
            params = {'snapshotDate': self.snapshot_time}

        params = urllib.parse.urlencode(params)
        url = '%s%s?%s' % (self.url_base, path, params)

        start_time = time.time()
        request = urllib.request.Request(url)
        response = urllib.request.urlopen(request)
        json_result = json.load(response)

        logger.info('Ran snapshots. %d seconds. URL=%s' % (time.time() - start_time, url))

        return json_result

    def balances(self, accounts, from_date=None, to_date=None, with_counterparties=None, excluding_counterparties=None, excluding_contra_accounts=None, with_tags=None, excluding_tags=None):
        from_date = None if from_date == '2000-01-01' else from_date
        account_balances = self.__get('/balances', {
            'accounts': ','.join(accounts),
            'from': from_date,
            'to': to_date,
            'withCounterparties': ','.join(with_counterparties or []),
            'excludingCounterparties': ','.join(excluding_counterparties or []),
            'excludingContraAccounts': ','.join(excluding_contra_accounts or []),
            'withTags': ','.join(with_tags or []),
            'excludingTags': ','.join(excluding_tags or []),
        })
        return account_balances

    """
    def transactions(self, accounts, from_date=None, to_date=None, chunk_frequency=None, with_counterparties=None, excluding_counterparties=None, excluding_contra_accounts=None):
        from_date = None if from_date == '2000-01-01' else from_date
        transaction_history = self.__get('/transactions', {
            'accounts': ','.join(accounts),
            'from': from_date,
            'to': to_date,
            'chunkFrequency': chunk_frequency,
            'withCounterparties': ','.join(with_counterparties or []),
            'excludingCounterparties': ','.join(excluding_counterparties or []),
            'excludingContraAccounts': ','.join(excluding_contra_accounts or [])
        })
        return transaction_history

    def get_transaction(self, transaction_id):
        return self.__get('/transaction/%s' % transaction_id)
    """
    
    @staticmethod
    def __accountifie_svc_url():
        return api_func('environment', 'variable', 'ACCOUNTIFIE_SVC_URL')




class Band(object):
    "Specifies what to fetch and create"
    def __init__(self, id=None, indent=0, group=None):
        self.id = id
        self.indent = indent
        self.group = group

    def get_rows(self, report):
        "Generate one or more rows. Base class does one row"

        row = dict(id=self.id, indent=self.indent)
        row['text'] = "Dummy report band"
        values = []
        for colspec in report.columns:
            values.append(Decimal('0.00'))
            row['values'] = values
        rows = [row]
        return rows


class TextBand(Band):
    """"Just inserts a slice of text.

    Suggested classes:
        h1 = bigger bolder heading
        h2 = slightly less prominent heading
        None, plain = ordinary text suitable for a paragraph
        small = smaller text for notes

        major = matches text of major band row
        minor = matches minor band row

    Any indent level will increase the default indent
    """
    def __init__(self, text, css_class=None, indent = 0, group=None):
        self.text = text
        self.id = None
        self.css_class = css_class
        self.indent = indent
        self.group = group

    def get_rows(self, report):
        #just one row
        return [dict(
            id=self.id,
            group=self.group,
            text=self.text, 
            css_class=self.css_class, 
            indent=self.indent, 
            type='text', 
            values=[''] * len(report.columns)
            )]
        #don't bother with the cells, the format layer can handle missing ones.

class BasicBand(object):

    def __init__(self, text, id=None, css_class=None, indent = 0, line_above=0, line_below=0, group=None, values=None, num_cols=1, type='basic'):
        self.text = text
        self.id = id
        self.group = group
        self.css_class = css_class
        self.indent = indent
        self.line_above=line_above
        self.line_below = line_below
        self.type = type
        if values:
            self.values = values
        else:
            self.values = [{'text': '', 'link':''} for x in range(num_cols)]

    def get_rows(self, num_cols=1):
        #just one row
        return [dict(
            id=self.id,
            text=self.text,
            css_class=self.css_class,
            indent=self.indent,
            type=self.type,
            values=self.values
            )]



from dateutil.parser import parse

from django.contrib.auth.decorators import login_required

import accountifie.toolkit.utils as utils
from accountifie.query.query_manager import QueryManager


def download_ledger(company_id, qs):
    
    # GET DATE PARAMS from request
    from_dt = qs.get('from')
    if type(from_dt) == list:
    	from_dt = from_dt[0]

    if from_dt:
    	trans = QueryManager().transactions(company_id, from_date=parse(from_dt).date())	
    else:
    	trans = QueryManager().transactions(company_id)

    def _get_bmo_id(rec):
        id_parts = rec['id'].split('.')
        return '%s.%s' % (id_parts[0], id_parts[1])

    for tr in trans:
        tr.update({'bmo_id': _get_bmo_id(tr)})

    return sorted(trans, key=lambda x: x['bmo_id'])




from accountifie.reporting.models import ReportDef
from accountifie.reporting.serializers import ReportDefSerializer


def reportdef(qstring):
    qs = ReportDef.objects.all()
    return ReportDefSerializer(qs, many=True).data




from .models import Alias, Variable, Config
from multipledispatch import dispatch


def alias(name, qstring):
    try:
        alias = Alias.objects.get(name=name)
        return dict((k, getattr(alias, k)) for k in ['display_as', 'name'])
    except:
        return None

@dispatch(dict)
def variable(qstring):
    return list(Variable.objects.order_by('id').values())

@dispatch(str, dict)
def variable(key, qstring):
    try:
        return Variable.objects.get(key=key).value
    except:
        raise ValueError("Can't find system variable %s" % key)

def variable_list(key, qstring):
    try:
        return Variable.objects.get(key=key).value.split(',')
    except:
        raise ValueError("Can't find system variable %s" % key)


def config(name, qstring):
    try:
        return Config.objects.get(name=name).reporting
    except:
        raise ValueError("Can't find config variable %s" % key)


# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reporting', '0003_auto_20160720_2313'),
    ]

    operations = [
        migrations.AddField(
            model_name='metricentry',
            name='label',
            field=models.CharField(max_length=20, null=True, blank=True),
        ),
    ]



from django.conf.urls import url


from . import views


urlpatterns = [
    url(r'get_report/$', views.create_report),
    url(r'api/(?P<api_view>[_a-zA-Z0-9]+)/$', views.api),
    url(r'upload/(?P<file_type>[_a-zA-Z0-9]+)/$', views.upload_file),
    # reports
    url(r'reports/(?P<id>[_a-zA-Z0-9]+)/$', views.report),
    
    # transaction history
    url(r'history/(?P<type>[()_a-zA-Z0-9]+)/(?P<id>[()_a-zA-Z0-9]+)/$', views.history),
]



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Metric',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=50)),
            ],
        ),
        migrations.CreateModel(
            name='MetricEntry',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date', models.DateField(db_index=True)),
                ('balance', models.DecimalField(max_digits=20, decimal_places=2)),
                ('metric', models.ForeignKey(blank=True, to='reporting.Metric', null=True)),
            ],
        ),
        migrations.CreateModel(
            name='ReportDef',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=50)),
                ('path', models.CharField(max_length=200)),
            ],
            options={
                'db_table': 'reporting_reportdef',
            },
        ),
    ]



"""
Used to generate strategy objects for dealing with transactions in query_manager.

To get the default strategy: QueryManagerStrategyFactory().get()
To get a specific strategy:  QueryManagerStrategyFactory().get(strategy='remote')

Also used to compare strategy output. See test/query_manager.py for example.
"""

import difflib
import pandas as pd
import sys
import time
import re

from dateutil.parser import parse as parse_date
from decimal import Decimal
from django.conf import settings
from pandas.util.testing import assert_frame_equal
from .query_manager_remote_strategy import QueryManagerRemoteStrategy
from .query_manager_local_strategy import QueryManagerLocalStrategy
from .query_manager_test_strategy import QueryManagerTestStrategy
from .query_manager_snapshot_strategy import QueryManagerSnapshotStrategy
from accountifie.common.api import api_func


class QueryManagerStrategyFactory(object):
  def __init__(self):
    self.force_default_strategy = None
    self.strategy_singletons = {
      'remote': QueryManagerRemoteStrategy(),
      'local': QueryManagerLocalStrategy(),
      'test': QueryManagerTestStrategy(),
    }
    try:
      self.noop_mutating_functions = api_func('environment', 'variable', 'DISABLE_ACCOUNTIFIE_SVC_MUTATES') == 'true'
    except:
      self.noop_mutating_functions = False

  def override_default_strategy(self, strategy):
    self.force_default_strategy = strategy

  def get(self, strategy=None):
    strategy = strategy or self.__default_strategy()
    if strategy in self.strategy_singletons:
      return self.strategy_singletons[strategy]

    if strategy == 'snapshot':
      return QueryManagerSnapshotStrategy()

    matched_snapshot_strategy = re.match(r'snapshot_(.+)', strategy)
    if matched_snapshot_strategy:
      date_str = matched_snapshot_strategy.group(1)
      snapshot_strategy = QueryManagerSnapshotStrategy()
      date = parse_date(date_str)
      snapshot_strategy.set_cache(date)
      return snapshot_strategy

    raise Exception('Unknown GL strategy: %s. Valid options: [remote, local, snapshot_DATESTRING].' % strategy)

  def get_transaction(self, company_id, transaction_id):
    if not self.noop_mutating_functions:
      for strategy in self.strategy_singletons:
        self.strategy_singletons[strategy].get_transaction(company_id, transaction_id)

  def upsert_transaction(self, transaction):
    if not self.noop_mutating_functions:
      for strategy in self.strategy_singletons:
        self.strategy_singletons[strategy].upsert_transaction(transaction)

  def delete_transaction(self, company_id, transaction_id):
    if not self.noop_mutating_functions:
      for strategy in self.strategy_singletons:
        self.strategy_singletons[strategy].delete_transaction(company_id, transaction_id)

  def delete_bmo_transactions(self, company_id, bmo_id):
    if not self.noop_mutating_functions:
      for strategy in self.strategy_singletons:
        self.strategy_singletons[strategy].delete_bmo_transactions(company_id, bmo_id)

  def erase(self, company_id):
    if not self.noop_mutating_functions:
      for strategy in self.strategy_singletons:
        self.strategy_singletons[strategy].erase(company_id)

  def set_fast_inserts(self, company_id, value):
    if not self.noop_mutating_functions:
      for strategy in self.strategy_singletons:
        self.strategy_singletons[strategy].set_fast_inserts(company_id, value)

  def take_snapshot(self, company_id):
    if not self.noop_mutating_functions:
      for strategy in self.strategy_singletons:
        self.strategy_singletons[strategy].take_snapshot(company_id)

  def compare(self, method_name, method_args):
    fetch_with = lambda obj: getattr(obj, method_name)(**method_args)
    fmt = lambda obj: ', '.join(['%s=%s' % (k, obj[k]) for k in obj])
    column_weights = { 'id': '000', 'date': '001' }
    weigh_column = lambda col: column_weights[col] + col if col in column_weights else '500' + col
    weighted_col_sort = lambda cols: sorted(cols, key=weigh_column)

    remote_start = time.time()
    remote_result = fetch_with(self.get(strategy='remote'))
    remote_time = time.time() - remote_start

    local_start = time.time()
    local_result  = fetch_with(self.get(strategy='local'))
    local_time = time.time() - local_start

    remote_pd = pd.DataFrame(remote_result).fillna(0)
    local_pd =  pd.DataFrame(local_result).fillna(0)

    is_equal = True
    match_error = ''
    try:
      assert_frame_equal(remote_pd, local_pd)
    except:
      match_error = sys.exc_info()[1]
      is_equal = False

    remote_res_str = remote_pd[weighted_col_sort(remote_pd.columns.values)].replace(Decimal(0), Decimal('0.00')).to_csv(sep=',', index=False, header=True).splitlines()
    local_res_str  = local_pd[weighted_col_sort(local_pd.columns.values)].replace(Decimal(0), Decimal('0.00')).to_csv(sep=',', index=False, header=True).splitlines()

    # Header row hack to ensure stays up top
    remote_res_str[0] = '!%s' % remote_res_str[0]
    local_res_str[0] = '!%s' % local_res_str[0]

    remote_res_str.sort()
    local_res_str.sort()

    # diff = difflib.unified_diff(remote_res_str, local_res_str, fromfile='remote_strategy', tofile='local_strategy', n=3)
    diff = difflib.ndiff(remote_res_str, local_res_str)

    print('=== DIFF ===')
    for line in diff:
      print(line)
    print('')

    print('!! Call:    %s(%s)' % (method_name, fmt(method_args)))
    print('!! Matches: %s' % (is_equal))
    if match_error: print('!! Match Error: %s' % match_error)
    print('!! Times:   local %0.2fs, remote %0.2fs' % (local_time, remote_time))
    print('')

  def __default_strategy(self):
    if self.force_default_strategy:
      return self.force_default_strategy

    try:
      return api_func('environment', 'variable', 'DEFAULT_GL_STRATEGY')
    except:
      return settings.DEFAULT_GL_STRATEGY


# in lieu of a cleaner python implementation of singleton pattern
instance = None
def getInstance():
  global instance
  if not instance:
    instance = QueryManagerStrategyFactory()
  return instance



"""
Adapted with permission from ReportLab's DocEngine framework
"""

import os
import json
from io import StringIO

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.views.generic.detail import DetailView
from django.shortcuts import render
from django.template import RequestContext
from django.contrib import messages
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.db import transaction
from django.apps import apps
from django.core.serializers.json import DjangoJSONEncoder

from accountifie.toolkit.utils import get_default_company, csv_to_modelattr, create_instance, \
                                      dirty_key, get_company, utcnow

from accountifie.toolkit.forms import FileForm


import logging

logger = logging.getLogger('default')


DATA_ROOT = getattr(settings, 'DATA_DIR', os.path.join(settings.ENVIRON_DIR, 'data'))
INCOMING_ROOT = os.path.join(DATA_ROOT, 'incoming')
PROCESSED_ROOT = os.path.join(DATA_ROOT, 'processed')


@login_required
def upload_file(request, config):
    file_type = config['file_type']

    context = {'file_type': file_type,}     
    form = FileForm(request.POST, request.FILES)

    if form.is_valid():
        upload = list(request.FILES.values())[0]
        file_name_with_timestamp = save_file(upload)
        company = get_company(request)

        model = config.pop('model')
        unique = config.pop('unique')
        name_cleaner = config.pop('name_cleaner')
        value_cleaner = config.pop('value_cleaner')
        exclude = config.pop('exclude')
        post_process = config.pop('post_process')

        result = process_incoming_file(model, unique, name_cleaner, value_cleaner, exclude, post_process,
                                                 file_type=config['file_type'],
                                                 file_name=file_name_with_timestamp,
                                                 company=company,
                                                 task_title="%s data for %s" % (file_type, company)
                                                 )

        output = dict((k, result.get(k)) for k in ['saved', 'key_errors', 'value_errors', 'found', 'data'])
        output['dupes'] = len(result.get('dups', {}))
        context = {'data': json.dumps(output, cls=DjangoJSONEncoder, indent=2)}
        context['title'] = 'File Upload Results'
        msg = 'Loaded %s. Checked %d -- %d new, %d dupes.' % (config['file_type'],
                                                              output['found'],
                                                              len(output['saved']),
                                                              output['dupes'])
        messages.info(request, msg)
        return render(request, 'api_display.html', context)
    else:
        context['error'] = 'form not in valid format'
        context['title'] = 'File Upload Results'
        return render(request, 'api_display.html', context)


def save_file(f):
    new_name = '%s_%s' % (utcnow().strftime('%Y%m%d%H%M%S'), f._name)
    if not os.path.exists(INCOMING_ROOT):
        os.makedirs(INCOMING_ROOT)
    full_path = os.path.join(INCOMING_ROOT, new_name)
    with open(full_path, 'wb') as destination:
        for chunk in f.chunks():
            destination.write(chunk)

    return new_name 


def process_incoming_file(model, unique, name_cleaner, value_cleaner, exclude, post_process, **config):
    out = StringIO()
    file_name = config.pop('file_name')
    context = {'file_name': file_name}
    file_type = config.pop('file_type')
    
    incoming_name = os.path.join(INCOMING_ROOT, file_name)
    
    with open(incoming_name, 'U') as f:

        data = json.loads(load_file(file=f, stdout=out, model=model, unique=unique, name_cleaner=name_cleaner,
                                    value_cleaner=value_cleaner, exclude=exclude, post_process=post_process, **config))

    context['found'] = sum([len(data[k]) for k in data])
    context['data'] = data['saved']
    if data['saved']:
        context['example'] = data['saved'][0]
    elif data['key_errors']:
        context['example'] = data['key_errors']
    else:
        context['example'] = dict(no_keys_found='no values found')
    for k in [x for x in list(data.keys()) if x not in list(context.keys())]:
        context[k] = data[k]
    
    post_processing(file_type)
    
    # move the file into 'processed' directory
    if not os.path.exists(PROCESSED_ROOT):
        os.makedirs(PROCESSED_ROOT)
    processed_name = os.path.join(PROCESSED_ROOT, file_name)
    os.rename(incoming_name, processed_name)
    return context


def post_processing(file_type):
    """
    For example matching counterparties based on historical patterns
    """
    return


def load_file(**config):
    file = config.pop('file', None)
    company = config.get('company', get_default_company())
    model = config.get('model')
    unique = config.get('unique')
    name_cleaner = config.get('name_cleaner')
    value_cleaner = config.get('value_cleaner')
    exclude = config.get('exclude', [])
    data = csv_to_modelattr(file, name_cleaner=clean_csv_fields, company=company)
    #with transaction.atomic():
    results = save_data(data, company, model, unique, name_cleaner, value_cleaner, exclude)
        
    return json.dumps(results, ensure_ascii=False)
    
                
def save_data(data, company, model, unique, name_cleaner, value_cleaner, exclude):
    if name_cleaner == None:
        name_cleaner = lambda name: name
    if value_cleaner == None:
        value_cleaner = lambda name, value: value
    if unique == None:
        unique = lambda instance: instance
    model = apps.get_model(*model.split('.'))

    dups = []
    saved = []
    value_errors = []

    for index, row in enumerate(data):
        dirty = dirty_key(row, model, name_cleaner=name_cleaner, value_cleaner=value_cleaner)
        
        if dirty:
            #if a key is wrong the whole file will fail anyway
            return dict(dups=dups, saved=saved, key_errors=dirty, value_errors=value_errors)
        with transaction.atomic():
            unique_instance = create_instance(row, model, name_cleaner=name_cleaner, value_cleaner=value_cleaner,
                                              unique=unique, company=company, exclude=exclude)
        if unique_instance:
            try:
                unique_instance.save()
            except Exception as e:
                print('exception right here', e)
                message = getattr(e, 'messages', [row])[0]
                value_errors.append((e.__class__.__name__, index, message))
                continue
            
            #need to deconstruct the model instance for later json serialization
            saved.append(dict([(k, v)for k,v in list(unique_instance.__dict__.items()) if k in [f.name for f in unique_instance._meta.fields]]))
        else:
            decoded = dict((k.decode('utf-8', 'ignore'),v.decode('utf-8', 'ignore')) for k,v in row.items())
            dups.append(decoded)
    
    return dict(dups=dups, saved=saved, key_errors=[], value_errors=value_errors)


def clean_csv_fields(name):
    return name.lower().replace(' ', '_').replace('-', '_').replace('check_or_slip_#', 'check_or_slip')
    


import os
import sys
import traceback
from . import csv

from django.conf import settings
from django.contrib import messages
from django.shortcuts import render
from django.http import HttpResponseRedirect, JsonResponse

import accountifie.common.uploaders
from accountifie.toolkit.forms import FileForm, LabelledFileForm
from accountifie.toolkit.utils import utcnow, get_default_company

import logging
logger = logging.getLogger('default')


DATA_ROOT = getattr(settings, 'DATA_DIR', os.path.join(settings.ENVIRON_DIR, 'data'))
INCOMING_ROOT = os.path.join(DATA_ROOT, 'incoming')
PROCESSED_ROOT = os.path.join(DATA_ROOT, 'processed')


def save_file(infile):
    new_name = '%s_%s' % (utcnow().strftime('%Y%m%d%H%M%S'), infile._name)
    if not os.path.exists(INCOMING_ROOT):
        os.makedirs(INCOMING_ROOT)
    full_path = os.path.join(INCOMING_ROOT, new_name)
    with open(full_path, 'wb') as destination:
        for chunk in infile.chunks():
            destination.write(chunk)
    return new_name 


def order_upload(request, processor, redirect_url=None, label=False, file_type='csv upload'):
    if request.method == 'POST':
        
        if label:
            form = LabelledFileForm(request.POST, request.FILES)
        else:
            form = FileForm(request.POST, request.FILES)

        if form.is_valid():
            upload = list(request.FILES.values())[0]
            file_name_with_timestamp = save_file(upload)

            if label:
                summary_msg, error_msgs = processor(file_name_with_timestamp,
                                                    label=form.cleaned_data['label'])
            else:
                summary_msg, error_msgs = processor(file_name_with_timestamp)
            
            messages.success(request, summary_msg)
            for err in error_msgs:
                messages.error(request, err)
        else:
            try:
                upload = list(request.FILES.values())[0]
                file_name_with_timestamp = save_file(upload)
                if label:
                    summary_msg, error_msgs = processor(file_name_with_timestamp,
                                                        label=form.cleaned_data['label'])
                else:
                    summary_msg, error_msgs = processor(file_name_with_timestamp)
                
                return JsonResponse({'summary': summary_msg, 'errors': error_msgs})
            except:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
                traceback.print_exc()
                msg = 'Could not process the file provided, please see below'
                return JsonResponse({'summary': msg, 'errors': []})

        if redirect_url:
            return HttpResponseRedirect(redirect_url)
        else:
            return HttpResponseRedirect(request.META.get('HTTP_REFERER'))
    else:
        form = FileForm()
        context = {'form': form, 'file_type': file_type}
        return render(request, 'common/upload_csv.html', context)


def csv_to_modelattr(open_file, name_cleaner=None, company=get_default_company()):
    '''takes the fields and values in the CSV and transforms them into a list of dicts where the keys
     will match model attributes. for example Full Description becomes full_description'''
    if name_cleaner == None:
        name_cleaner = lambda name: name
    f_csv = csv.DictReader(open_file) 
    csv_to_modelattr = dict([(name, name_cleaner(name)) for name in f_csv.fieldnames])
    csv_to_modelattr['company_id'] = company

    return [dict([(csv_to_modelattr[name], value) for name, value in list(row.items()) if name in csv_to_modelattr]) for row in f_csv]



def get_foreignkeys(model):
    return dict(((f.name, f.rel.to) for f in model._meta.fields if f.__class__ == models.ForeignKey))

def get_fk_attr(model):
    return [f.name for f in model._meta.fields if f.__class__ == models.ForeignKey]

def get_pk_name(model):
    return model._meta.pk.name
    
def instance_nonrel_data(row, model, name_cleaner=None, value_cleaner=None):
    model_flds =  model._meta.get_all_field_names()
    instance_data_no_fk = dict((name_cleaner(name), value_cleaner(name, value)) for name, value in list(row.items()) if name_cleaner(name)
                        and name_cleaner(name) not in get_fk_attr(model) and name_cleaner(name) in model_flds)
    return model(**instance_data_no_fk)
    
def set_foreignkeys(instance, row, model, name_cleaner=None, value_cleaner=None):
    if get_foreignkeys(model): 
        instance_fk = dict((name_cleaner(name), value_cleaner(name, value)) for name, value in list(row.items()) if name_cleaner(name) 
                        and name_cleaner(name) in get_fk_attr(model))
        for fk in list(get_foreignkeys(model).items()):
            if fk[0] in instance_fk:
                try:
                    related = fk[1].objects.get(pk=instance_fk[fk[0]])
                    setattr(instance, fk[0], related)
                except:
                    logger.error("No ForeignKey %s %s.  %s" % (fk[0], str(fk[1]), instance_fk))
    return instance
    
def dirty_key(row, model=None, unique=None, name_cleaner=None, value_cleaner=None):
    dirty = [name_cleaner(k) for k in list(row.keys()) 
                if name_cleaner(k) not in [f.name for f in [field for field in model._meta.fields 
                        if field not in get_fk_attr(model)]] 
                                if name_cleaner(k)]

    return dirty
    
def create_instance(row, model, name_cleaner=None, value_cleaner=None, unique=None, exclude=[], company=get_default_company()):
    row['company'] = company
    non_rel_instance = instance_nonrel_data(row, model, name_cleaner=name_cleaner, value_cleaner=value_cleaner)
    if non_rel_instance.id in exclude:
        return None

    full_instance = set_foreignkeys(non_rel_instance, row, model, name_cleaner=name_cleaner, value_cleaner=value_cleaner)    

    return unique(full_instance)
    



"""
Adapted with permission from ReportLab's DocEngine framework
"""


from django.apps import apps
from django.db import models
from django.conf import settings


from django.apps import apps
def all_concrete_models():
    R = [].append
    M = apps.get_models()
    for app_label, models in apps.all_models.items():
        models = [model for name,model in models.items() if model in M and not model._meta.abstract]
        if models:
            R((app_label,models))
    return R.__self__



def field_is_relational(field):
    relational_fields = [models.ForeignKey, models.ManyToManyField, models.OneToOneField]
    return [rel_field for rel_field in relational_fields if isinstance(field, rel_field)]

def relational_fields(model):
    return [field for field in model._meta.fields if field_is_relational(field)]

def all_models_with_relation():
    return [(app, [model for model in models if relational_fields(model)]) for app, models in all_concrete_models()]

def all_relations():
    return [(app, [[model, [(field, field.rel.to) for field in relational_fields(model)]]
                                                    for model in models])
                                                        for app, models in all_models_with_relation() if models]





def column_chart(title="Column Chart", y_label="y-Axis", data_labels=False, colored_cols=False):
    chart_data = {
        'chart': {'type': 'column'},
        'yAxis': {},
        'plotOptions': {'series': {}},
        'credits': {'enabled': False}
        }

    chart_data['title'] = title
    chart_data['yAxis']['title'] = {'text': y_label}
    if data_labels is True:
        chart_data['plotOptions']['series']['dataLabels'] = {'enabled': True}
    if colored_cols is True: 
        chart_data['plotOptions']['series']['colorByPoint'] = True

    return chart_data

def line_chart(title="Column Chart", y_label="y-Axis", data_labels=False):
    chart_data = {
        'chart': {'type': 'line'},
        'yAxis': {},
        'plotOptions': {'series': {}},
        'credits': {'enabled': False}
        }

    chart_data['title'] = title
    chart_data['yAxis']['title'] = {'text': y_label}
    if data_labels is True:
        chart_data['plotOptions']['series']['dataLabels'] = {'enabled': True}
    
    return chart_data



import os
import csv
import datetime
from dateutil.parser import parse
from decimal import Decimal

from django.conf import settings
from django.contrib import messages
from django.shortcuts import render
from django.template import RequestContext
from django.http import HttpResponseRedirect


from accountifie.toolkit.forms import LabelledFileForm
from accountifie.common.uploaders.csv import save_file
from .models import Metric, MetricEntry

DATA_ROOT = getattr(settings, 'DATA_DIR', os.path.join(settings.ENVIRON_DIR, 'data'))
INCOMING_ROOT = os.path.join(DATA_ROOT, 'incoming')
PROCESSED_ROOT = os.path.join(DATA_ROOT, 'processed')


def process_metrics(file_name, label=None):
    incoming_name = os.path.join(INCOMING_ROOT, file_name)
    with open(incoming_name, 'U') as f:
        reader = csv.reader(f)
        header = next(reader)
        rows = []
        for row in reader:
            rows.append(row)

    # validation

    new_metrics = 0
    metric_objs = {}
    if header[0] != 'DATE':
        return {'error': 'First column must be DATE'}

    for col in header[1:]:
        metric_obj = Metric.objects.filter(name=col).first()
        if not metric_obj:
            metric_obj = Metric(name=col)
            metric_obj.save()
            new_metrics += 1
        metric_objs[col] = metric_obj

    as_of = datetime.datetime.now().date()
    new_entry_cnt = 0
    update_entry_cnt = 0

    for row in rows:

        try:
            dt = parse(row[0])
        except:
            return {'error': 'Bad DATE: %s' % row[0]}
        for col_num in range(1, len(header)):
            try:
                value = Decimal(row[col_num])
                col = header[col_num]
                m_entry = {}
                m_entry['metric'] = metric_objs[col]
                m_entry['date'] = dt
                m_entry['balance'] = Decimal(row[col_num])
                m_entry['label'] = label
                m_entry['as_of'] = as_of
                obj = MetricEntry.objects \
                                 .filter(metric=m_entry['metric'],
                                         date=dt,
                                         label=label) \
                                 .first()
                if obj:
                    obj.balance = Decimal(row[col_num])
                    obj.as_of = as_of
                    update_entry_cnt += 1
                else:
                    MetricEntry(**m_entry).save()
                    new_entry_cnt += 1
            except:
                pass

    msg_list = []
    msg_list.append('Created %d metric entries' % new_entry_cnt)
    msg_list.append('Updated %d metric entries' % new_metrics)
    msg_list.append('%d new metrics' % update_entry_cnt)
    return msg_list, []



import datetime
import mock
from django.test import TestCase

from accountifie.gl.models import Account, Company, Counterparty
from accountifie.reporting.core_reports.trialbalance import TrialBalance
from accountifie.query.query_manager_strategy import QueryManagerStrategy

class TrialBalanceTestCase(TestCase):
  def setUp(self):
    self.default_company = Company(name='TEST_COMPANY', id='TEST')
    self.default_company.save()

    Account(id='1001', path='test.one', display_name='TestOne', role='asset').save()
    Account(id='1002', path='test.two', display_name='TestTwo', role='expense').save()
    Account(id='1003', path='test.three', display_name='TestThree', role='liability').save()

  def test_trial_balance_happy_path(self):
    # setup mock strategy
    mock_strategy = mock.create_autospec(QueryManagerStrategy, instance=True)
    mock_strategy.account_balances_for_dates.return_value = {
      'balance': {
        '1001': 11.00,
        '1002': 22.00,
        '1003': 33.00
      }
    }

    # create trial_balance
    trial_balance = TrialBalance('TEST')
    trial_balance.set_gl_strategy(mock_strategy)

    # run calc
    result = trial_balance.calcs()

    # test calls
    mock_strategy.account_balances_for_dates.assert_called_once_with(
      company_id='TEST',
      account_ids=['1001', '1002', '1003'],
      # TODO: remove hardcoded start date
      dates={'balance': {'start': datetime.date(2013, 1, 1), 'end': None}},
      with_counterparties=None,
      excl_interco=False,
      excl_contra=None,
      with_tags=None,
      excl_tags=None
    )

    # test result
    self.assertEqual(result[0]['label'], '1001: TestOne')
    self.assertEqual(result[0]['Credits']['text'], 0.0)
    self.assertEqual(result[0]['Debits']['text'], 11.0)

    self.assertEqual(result[1]['label'], '1002: TestTwo')
    self.assertEqual(result[1]['Credits']['text'], 0.0)
    self.assertEqual(result[1]['Debits']['text'], 22.0)

    self.assertEqual(result[2]['label'], '1003: TestThree')
    self.assertEqual(result[2]['Credits']['text'], -33.0)
    self.assertEqual(result[2]['Debits']['text'], 0.0)

    self.assertEqual(result[3]['label'], 'Totals')
    self.assertEqual(result[3]['Credits']['text'], -33.0)
    self.assertEqual(result[3]['Debits']['text'], 33.0)



import json
from dateutil.parser import parse

from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.core.serializers.json import DjangoJSONEncoder

import accountifie.reporting.api

import logging
logger = logging.getLogger('default')


def api(request, api_view):
    params = request.GET
    return HttpResponse(json.dumps(accountifie.reporting.api.get(api_view, params), cls=DjangoJSONEncoder), content_type="application/json")

def balance_trends(request):
    dt = parse(request.GET.get('date'))
    acct_list = request.GET.get('acct_list').split('.')

    return HttpResponse(json.dumps(accountifie.reporting.api.balance_trends(dt, acct_list), cls=DjangoJSONEncoder), content_type="application/json")



"""
Based on Andy Robinson's DoubleTalk accounting framework; used
with permission
"""

from decimal import Decimal
import logging

from django.db import transaction

import accountifie.query.query_manager_strategy_factory as QMSF
from accountifie.gl.models import TranLine


DZERO = Decimal('0')

logger = logging.getLogger('default')

GL_STRATEGY = 'local'


def _model_to_id(x):
    return x if type(x) == str else x.id


class BusinessModelObject(object):
    """Something else which has an effect on the GL.

    BMOs will typically create General Ledger entries when
    saved.  They should efficiently update/change their
    GL entries and not create database churn unless there
    is a change.

    Make other models inherit from this as a mixin.

    Regrettably one then needs to override save() and
    delete() for the other models, until I find a cunning
    way to make it happen.

    """

    def delete_from_gl(self):
        "Find and remove all GL entries which depend on self"
        bmo_id = '%s.%s' %(self.short_code, self.id)
        QMSF.getInstance().get().delete_bmo_transactions(self.company.id, bmo_id)


    def create_gl_transactions(self, trans):
        "Make any GL transactions this needs"
        for td in trans:
            try:
                comment = td['comment'][:99]
            except:
                comment = ''

            # old style
            if GL_STRATEGY == 'local':
                print('=' * 20)
                print(td['lines'][0])
                lines = [{'company_id': _model_to_id(td['company']),
                          'date': td['date'],
                          'date_end': td.get('date_end', td['date']),
                          'comment': comment,
                          'account_id': _model_to_id(account),
                          'amount': Decimal("{0:.2f}".format(amount)),
                          'counterparty_id': _model_to_id(counterparty),
                          'tags': ','.join(tags or []),
                          'bmo_id': td['bmo_id'],
                          'source_object': self
                          } for account, amount, counterparty, tags in td['lines']]
                if sum(l['amount'] for l in lines) == DZERO:
                    # 1. check for balance
                    # 2. check for changes

                    # 3. if any changes then just delete/set to historical
                    #    the prior ones
                    save_tranlines(lines)

                else:
                    logger.error('Imbalanced GL entries for %s' % td['bmo_id'])

            elif GL_STRATEGY == 'remote':
                source_object = {'object_id': self.id,
                                 'model_name': self._meta.model_name,
                                 'app_name': self._meta.app_label}
                td['source_object'] = source_object

                lines = [{'company': _model_to_id(td['company']),
                          'date': str(td['date']),
                          'date_end': str(td.get('date_end', None) or td['date']),
                          'comment': comment,
                          'account': _model_to_id(l['account']),
                          'amount': "{0:.2f}".format(l['amount']),
                          'counterparty': _model_to_id(l['counterparty']),
                          'tags': l.get('tags', None) or [],
                          'bmo_id': td['bmo_id'],
                          'source_object': self
                          } for account, amount, counterparty, tags in td['lines']]
                if sum(l['amount'] for l in lines) == DZERO:
                    QMSF.getInstance() \
                        .get(strategy=GL_STRATEGY) \
                        .create_gl_transactions(td, lines, td['bmo_id'], td['bmo_id'])
                else:
                    logger.error('Imbalanced GL entries for %s' % td['bmo_id'])

    def update_gl(self):
        "Fix up GL after any kind of change"
        trans = self.get_gl_transactions()
        self.delete_from_gl()
        self.create_gl_transactions(trans)

    def get_gl_transactions(self):
        """Override this.  Create a list of transactions.

        Create a list of transactions in dictionary form
        which should exist.  It will create them all in the
        GL on saving if they do not already exist.
        """
        return []

    def get_company(self):
        return self.company.id


@transaction.atomic
def save_tranlines(lines):
    for l in lines:
        TranLine(**l).save()


def on_bmo_save(sender, **kwargs):
    """Alternative method - maintain GL through a signal

    Master-detail BMOs can't use a simple save hook, and
    instead need to set a signal to call this'
    """
    sender.update_gl()



from multipledispatch import dispatch

from accountifie.gl.models import Account
from accountifie.common.api import api_wrapper
import logging

logger = logging.getLogger('default')


from django.conf import settings
import os


def filter_it(x, ex_list):
    for e in ex_list:
        if e in x:
            return False
    return True
    

def path_accounts(path, qstring={}):
    excl = qstring.get('excl', None)
    incl = qstring.get('incl', None)

    if excl:
        acct_list = [a for a in list(Account.objects.filter(path__contains=str(path)).values()) if filter_it(a['path'], excl.split(','))]
    elif incl:
        acct_list = [a for a in list(Account.objects.filter(path__contains=str(path)).values()) if not filter_it(a['path'], incl.split(','))]
    else:
        acct_list = list(Account.objects.filter(path__contains=str(path)).values())

    return acct_list


@dispatch(dict)
def account(qstring):
    return list(Account.objects.order_by('id').values())


@dispatch(str, dict)
def account(acct_id, qstring):
    try:
        acct = (list(Account.objects.filter(pk=acct_id)) + list(Account.objects.filter(path=acct_id)))[0]
        flds = ['display_name', 'role', 'path', 'id']
        return dict((k,v) for k,v in acct.__dict__.items() if k in flds)
    except:
        return None


def child_paths(path, qstring={}):
    path_len = len(path.split('.'))
    all_sub_paths = [x.path for x in Account.objects.filter(path__contains=path) if len(x.path.split('.')) > path_len]
    all_child_paths = ['.'.join(x.split('.')[:(path_len+1)]) for x in all_sub_paths]
    return list(set(all_child_paths))




from django.db import models

class Variable(models.Model):
  key = models.CharField(max_length=100, unique=True)
  value = models.CharField(max_length=100)

  def __unicode__(self):
    return self.key


class Config(models.Model):
  name = models.CharField(max_length=20, unique=True)
  reporting = models.ForeignKey(Variable)

class Alias(models.Model):
  name = models.CharField(max_length=100)
  display_as = models.CharField(max_length=100)
  context = models.ForeignKey('gl.Company', null=True, blank=True)

def get_env_variable(key):
  try:
    return Variable.objects.get(key=key).value
  except:
    raise ValueError("Can't find system variable %s" %key)

def get_env_alias(name):
  try:
    return Alias.objects.get(name=name).to_dict()
  except:
    raise ValueError("Can't find alias variable %s" %key)

def get_env_config(key):
  try:
    return Config.objects.get(name=name).reporting.value
  except:
    raise ValueError("Can't find config variable %s" %key)


"""
Adapted with permission from ReportLab's DocEngine framework
"""




from django.conf import settings
from django.conf.urls import include, url, handler500
from django.core.urlresolvers import reverse_lazy
from django.template.loader import select_template
from django.views.generic.base import RedirectView, TemplateView

import accountifie.common.views
import accountifie.common.api

if 'docengine.register' in settings.INSTALLED_APPS:
    if getattr(settings, 'REGISTER_ONE_STEP', False):
        _ACCOUNT_URLS = 'docengine.register.simple_urls'
    else:
        _ACCOUNT_URLS = 'docengine.register.urls'
else:
    _ACCOUNT_URLS = 'accountifie.common.auth_urls'


handler500 = 'accountifie.common.views.custom_500'

urlpatterns = [
    url(r'api/(?P<group>[_a-zA-Z0-9]+)/(?P<resource>[_a-zA-Z0-9]+)/(?P<item>(.+))/$', accountifie.common.api.get_item),
    url(r'api/(?P<group>[_a-zA-Z0-9]+)/(?P<resource>[_a-zA-Z0-9]+)/$', accountifie.common.api.get_resource),
    url(r'chart/(?P<group>[_a-zA-Z0-9]+)/(?P<chart>[_a-zA-Z0-9]+)/$', accountifie.common.api.get_chart),

    url(r'^test-7491/$', accountifie.common.views.test7491, name='health-check'),
    url(r'^tests/error/$', accountifie.common.views.deliberateError, name='deliberate error'),
    url(r'^api/upload/$', accountifie.common.views.upload),
    url(r'^dashboard/', include('accountifie.dashboard.urls'), name="docengine_dashboard"),
    url(r'^docs/$', TemplateView.as_view(template_name='common/docs.html')),
    url(r'^about/$', TemplateView.as_view(template_name='common/about.html')),
    url(r'', include(_ACCOUNT_URLS)),
    url(r'^$', accountifie.common.views.index, name='index'),
    url(r'^check_all_tasks/$', accountifie.common.views.check_all_tasks, name="check_all_tasks"),
    # shortcuts
    #url(r'^login/', RedirectView.as_view(url=reverse_lazy("accounts_login"))),
    #url(r'^logout/', RedirectView.as_view(url=reverse_lazy("accounts_logout"))),
]



import csv
import json

from django.http import HttpResponse
from django.shortcuts import render
from django.template import RequestContext, Context
from django.core.serializers.json import DjangoJSONEncoder


import accountifie.reporting.rptutils as rptutils
import accountifie.toolkit.utils as utils

import logging

logger = logging.getLogger('default')


def report(request, id):

    report, is_report = rptutils.report_prep(request, id)
    if not is_report:
            return report
        
    if report.from_file:
        try:
            report_data = json.loads(report.from_file)
        except:
            msg = "Sorry. file source is not recognised : %s" % report.from_file
            return render(request, 'rpt_doesnt_exist.html', {'message': msg})
    else:
        report_data = report.calcs()

    if report.format == 'json':
        return HttpResponse(json.dumps(report_data, cls=DjangoJSONEncoder), content_type="application/json")
    elif report.format == 'csv':
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename="report.csv"'

        writer = csv.writer(response)
        writer.writerow([''] + report.column_order)
        map_values = lambda x: '' if x == '' else str(x['text']).replace('(', '-').replace(')', '')
        
        for row in report_data:
            if row['fmt_tag'] != 'header':
                writer.writerow([row['label']] + [map_values(row[col]) for col in report.column_order])
        return response
    elif report.format == 'html':
        # not bootstrap ... the more custom style for more complex reports
        context = report.html_report_context()
        context['query_string'] = request.META['QUERY_STRING']
        context['rows'] = []
        for rec in report_data:
            context['rows'] += report.get_row(rec)

        return render(request, 'report.html', context)
    else:
        msg = "Sorry. This format is not recognised : %s" % format
        return render(request, '404.html', {'message': msg}), False



# -*- coding: utf-8 -*-


from django.db import migrations, models
import datetime


class Migration(migrations.Migration):

    dependencies = [
        ('reporting', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='metricentry',
            name='as_of',
            field=models.DateField(default=datetime.date(2016, 7, 19), db_index=True),
        ),
    ]



"""Mini-framework for generating banded reports


A Report is initialised with a list of time identifiers for the columns,
and a list of bands running vertically.   A Band can be a piece of text,
or a row for an account, or a group of rows involving some kind of
drill-down.

When executed, a report returns a list of rows.  Each row is a dictionary
containing information to be formatted.   The possible entries are...

    id:  optional. logical identifier to be used in calculations or as
        a target for links
    type: required. 'text' or 'account'

    values:  required for non-text rows.
        the numbers to be drawn in each numeric cell.  Reports do a last minute
        pass over all the values formatting them as strings (accounting format)

    css_class:  the main typographic class we will use for its text

    indent:  default 0, 1 or 2 indent it further across to allow nesting

    line_above:  False or True or 0 or 1 or 2 (for double)
    line_below:  False or True or 0 or 1 or 2 (for double)




"""
from dateutil.parser import parse
import datetime


from .bands import TextBand, BasicBand
from .reportdef import ReportDef

import accountifie.query.query_manager
from accountifie.common.api import api_func
import accountifie.toolkit.utils as utils
import accountifie.reporting.rptutils as rptutils




ITEM = {'css_class': 'minor', 'indent': 1, 'type': 'group_item'}
MINOR_TOTAL = {'css_class': 'minor', 'indent': 0, 'type': 'group_total'}
MAJOR_TOTAL = {'css_class': 'major', 'indent': 0, 'type': 'group_total2'}
WARNING = {'css_class': 'warning', 'indent': 0, 'type': 'normal'}
ROW_FORMATS = {'item': ITEM, 'minor_total': MINOR_TOTAL, 'major_total': MAJOR_TOTAL, 'warning': WARNING}



class Report(object):
    def __init__(self, company_id, description='', date=None, title='', subtitle='', footer='', bands=[], columns=None, calc_type='as_of'):
        self.description = description
        self.title = title
        self.subtitle = subtitle
        self.footer = footer
        self.dflt_title = title
        self.company_id = company_id
        self.bands = bands
        self.columns = columns
        self.calc_type = calc_type
        self.column_order = None
        self.date = None
        self.path = None
        self.set_company()
        self.by_id = {}
        self.format = 'html'
        self.from_file = None
        


    def set_company(self):
        self.company_name = api_func('gl', 'company', self.company_id)['name']


    def set_columns(self, columns, column_order=None):
        self.columns = columns
        self.column_order = column_order

    def set_gl_strategy(self, gl_strategy):
        self.query_manager = accountifie.query.query_manager.QueryManager(gl_strategy=gl_strategy)

    
    def configure(self, config):
        qs_matches = rptutils.qs_parse(config)

        if len(qs_matches) == 0:
            raise ValueError('Unexpected query string: %s' % repr(config))
        elif len(qs_matches) > 1:
            raise ValueError('Unexpected query string: %s' % repr(config))
        else:
            config['config_type'] = qs_matches[0]


        if config['config_type'] == 'shortcut':
            config.pop('config_type')
            config.update(rptutils.parse_shortcut(config['col_tag']))


        if config['config_type'] == 'date':
            dt = rptutils.date_from_shortcut(config['date'])
            config.update(rptutils.config_fromdate(self.calc_type, self.description, dt))
        elif config['config_type'] == 'period':
            config.update(rptutils.config_fromperiod(self.calc_type, self.description, config))
        elif config['config_type'] == 'date_range':
            config.update(rptutils.config_fromdaterange(self.calc_type, self.description, config))
        
        self.title = config.get('title')
        self.set_columns(config['columns'], column_order=config.get('column_order'))
        self.date = config.get('date')
        self.path = config.get('path')
        return


    def get_row(self, df_row):
        if 'fmt_tag' in df_row:
            if df_row['fmt_tag'] == 'header':
                return TextBand(df_row['label'], css_class='normal').get_rows(self)
            else:
                values = [utils.entry(df_row[col_label]['text'] , link=df_row[col_label]['link']) for col_label in self.column_order]
                fmt_tag = df_row['fmt_tag']
                _css_class = ROW_FORMATS[fmt_tag]['css_class']
                _indent = ROW_FORMATS[fmt_tag]['indent']
                _type = ROW_FORMATS[fmt_tag]['type']
                return BasicBand(df_row['label'], css_class=_css_class, values=values, indent=_indent, type=_type).get_rows(self)
        else:
            values = [utils.entry(df_row[col_label]['text'] , link=df_row[col_label]['link']) for col_label in self.column_order]
            return BasicBand(df_row['label'], css_class=df_row['css_class'], values=values, \
                                indent=df_row['indent'], type=df_row['type']).get_rows(self)


    def _context(self):
        if self.column_order:
            column_titles = self.column_order
        else:
            column_titles = list(self.columns.keys())

        columns = [self.columns[x] for x in column_titles]
        context = dict(
                report=self,
                columns=columns,
                column_titles=column_titles,
                title=self.title,
                colcount=2+len(self.columns),
                italic_styles=['group_header', 'group_total'],
                no_space_before_styles=['group_item', 'group_total'],
                numeric_column_indices=list(range(2, len(self.columns) + 2)),
                )

        return context


    def html_report_context(self):
        return self._context()

    def pdf_report_context(self):
        return self._context()

    def get_content(self):
        "Returns list of dictionaries with info to format"
        self.by_id = {}
        content = []

        for band in self.bands:
            this_content = band.get_rows(self)
            content.extend(this_content)

            #allow next few rows to have access to wwhat we worked out
            if band.id and this_content:
                last_row = this_content[-1]
                #last row has values for maths
                if 'values' in last_row:
                    self.by_id[band.id] = last_row['values']
        #at the end, walk through formatting
        for row in content:
            if 'values' in row:
                if row['values'] != '':
                    row['values'] = [{'text': utils.fmt(c['text']), 'link': c['link']} if c != '' else '' for c in row['values']]
                    
        return content



import json
import csv
import itertools
import logging
import importlib
import sys
import traceback
import flatdict

from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponse
from django.shortcuts import render
from django.template import RequestContext


logger = logging.getLogger('default')

def api_wrapper(func):
    def api_call(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:
            ex_type, ex, tb = sys.exc_info()
            logger.exception('error in api call: %s.%s --- %s' %(func.__module__, func.__name__, ex))
            return None
    return api_call


def resource_func(group, resource, qstring={}, chart=False):
    try:
        module_class = 'chart' if chart else 'apiv1'
        api_module = get_module('%s.%s' % (group, module_class))
        api_method = getattr(api_module, resource)
        api_call = api_wrapper(api_method)
        return api_call(qstring)
    except:
        logger.exception('Error calling api %s:%s' % (group, resource))
        return


def item_func(group, resource, item, qstring={}):
    try:
        api_module = get_module('%s.%s' % (group, 'apiv1'))
        api_method = getattr(api_module, resource)
        api_call = api_wrapper(api_method)
        return api_call(str(item), dict(qstring))
    except:
        logger.exception('Error calling api %s:%s:%s' % (group, resource, item))
        return


def get_module(group):
    try:
        return importlib.import_module(group)
    except:
        pass

    try:
        return importlib.import_module('accountifie.%s' % group)
    except:
        logger.exception("couldn't find api group %s" % group)
        return None

def api_func(*args, **kwargs):
    qs = kwargs.get('qstring', {})

    try:
        if len(args)==2:
            return resource_func(args[0], args[1], qstring=qs)
        elif len(args)==3:
            return item_func(args[0], args[1], args[2], qstring=qs)
        else:
            logger.exception('error calling api_func: %s' % str(args))
    except:
        logger.exception('error calling api_func: %s' % str(args))
        return


def get_resource(request, group, resource):
    qs = dict((k, request.GET.get(k)) for k in list(request.GET.keys()))
    raw = (qs.get('raw') == 'true')
    as_csv = (qs.get('as_csv') == 'true')
    if raw:
        return HttpResponse(json.dumps(resource_func(group, resource, qstring=qs), cls=DjangoJSONEncoder), content_type="application/json")
    elif as_csv:
        data = resource_func(group, resource, qstring=qs)
        return output_as_csv(list(data), label=resource)
    else:
        context = {'data': json.dumps(resource_func(group, resource, qstring=qs), cls=DjangoJSONEncoder, indent=2)}
        context['title'] = 'API call: /%s/%s' % (group, resource)
        return render(request, 'api_display.html', context)


def get_chart(request, group, chart):
    qs = dict((k,v) for k,v in request.GET.items())
    return HttpResponse(json.dumps(resource_func(group, chart, qstring=qs, chart=True),
                                   cls=DjangoJSONEncoder), content_type="application/json")


def output_as_csv(data, label='output'):
    if type(data) == list:
        flat_data = [flatdict.FlatDict(d) for d in data]
    else:
        flat_data = [flatdict.FlatDict(data)]

    response = HttpResponse(content_type='text/csv')
    file_name = '%s.csv' % label
    response['Content-Disposition'] = 'attachment; filename=%s' % file_name
    writer = csv.writer(response)

    cols = list(set(itertools.chain.from_iterable([list(x.as_dict().keys()) for x in flat_data])))
    
    writer.writerow(cols)
    for r in flat_data:
        line = [r.get(c, '') for c in cols]
        writer.writerow(line)
    return response



def get_item(request, group, resource, item):
    qs = dict((k, request.GET.get(k)) for k in list(request.GET.keys()))
    raw = (qs.get('raw') == 'true')
    as_csv = (qs.get('as_csv') == 'true')
    if raw:
        return HttpResponse(json.dumps(item_func(group, resource, item, qstring=qs), cls=DjangoJSONEncoder), content_type="application/json")
    elif as_csv:
        data = item_func(group, resource, item, qstring=qs)
        return output_as_csv(list(data), label='%s:%s' % (resource, item))
    else:
        context = {}
        context['data'] = json.dumps(item_func(group, resource, item, qstring=qs), cls=DjangoJSONEncoder, indent=2)
        context['title'] = 'API call: /%s/%s/%s' % (group, resource, item)
        return render(request, 'api_display.html', context)



from django.db import models

CALC_TYPES = [('as_of', 'Point in Time'),
			  ('diff', 'Change Over Time')]

class ReportDef(models.Model):
    name = models.CharField(max_length=50)
    path = models.CharField(max_length=200)
    description = models.CharField(max_length=100)
    calc_type = models.CharField(max_length=20, choices=CALC_TYPES)

    class Meta:
        app_label = 'reporting'
        db_table = 'reporting_reportdef'
    
    def __str__(self):
        return self.name




import json
from dateutil.parser import parse
import datetime

from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponseRedirect, Http404, HttpResponse
from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.template import RequestContext

import accountifie.toolkit.utils as utils
import accountifie.reporting.rptutils as rptutils
from accountifie.common.api import api_func
import accountifie.query.query_manager as QM


def get(api_view, params):
    return globals()[api_view](params)


def html_report_snippet(rpt_id, company_id, as_of=None, col_tag=None, path=None, version='v1', gl_strategy=None):
    config = {}
    if as_of:
        config['as_of'] = as_of
    if col_tag:
        config['col_tag'] = col_tag

    report = rptutils.get_report(rpt_id, company_id, version=version)

    report.configure(config)
    report.set_gl_strategy(gl_strategy)

    report_data = [x for x in report.calcs() if x['fmt_tag']!='header']

    context = report.html_report_context()
    context['rows'] = []
    for rec in report_data:
        context['rows'] += report.get_row(rec)

    return context

def balance_trends(params):
    dt = parse(params['date'])

    accounts = api_func('gl', 'account')
    if 'accts_path' in params:
        acct_list = [x['id'] for x in accounts if params['accts_path'] in x['path']]
    else:
        acct_list = params['acct_list'].split('.')
    
    gl_strategy = params.get('gl_strategy', None)
    company_id = params.get('company_id', utils.get_default_company())
    
    # 3 months for now
    report = rptutils.get_report('AccountActivity', company_id, version='v1')
    report.set_gl_strategy(gl_strategy)

    M_1 = utils.end_of_prev_month(dt.month, dt.year)
    M_2 = utils.end_of_prev_month(M_1.month, M_1.year)

    three_mth = {}
    three_mth['M_0'] = '%dM%s' % (dt.year, '{:02d}'.format(dt.month))
    three_mth['M_1'] = '%dM%s' % (M_1.year, '{:02d}'.format(M_1.month))
    three_mth['M_2'] = '%dM%s' % (M_2.year, '{:02d}'.format(M_2.month))

    three_mth_order = ['M_2','M_1','M_0']

    col_tag = '%dM%s' % (dt.year, '{:02d}'.format(dt.month))
    report.configure({'col_tag': col_tag})
    report.columns = three_mth
    report.column_order = three_mth_order
    report.acct_list = acct_list
    
    return [x for x in report.calcs() if x['fmt_tag']!='header']




from django.conf.urls import include, url
from django.contrib import admin

admin.autodiscover()

urlpatterns = [

    url(r'^admin/', include(admin.site.urls)),

    url(r'', include('accountifie.toolkit.urls')),
    url(r'', include('accountifie.common.urls')),
    url(r'', include('accountifie.celery.urls')),
    url(r'', include('accountifie.gl.urls')),
    url(r'', include('accountifie.snapshot.urls')),
    url(r'reporting/', include('accountifie.reporting.urls')),
]



import datetime
from datetime import tzinfo
from dateutil.parser import parse
import isoweek
import re


ZERO = datetime.timedelta(0)

# A UTC class.
class utc(tzinfo):
    """UTC"""
    def utcoffset(self, dt):
        return ZERO
    def tzname(self, dt):
        return "UTC"
    def dst(self, dt):
        return ZERO

UTC = utc()

def utcnow():
    u = datetime.datetime.utcnow()
    return u.replace(tzinfo=UTC)

def daterange(start, end, bus_days_only=True):
    dates = [start + datetime.timedelta(days=d) for d in range(0, (end - start).days + 1)]

    if bus_days_only:
        return [d for d in dates if d.weekday() < 5]
    else:
        return dates

def today():
    return datetime.datetime.now().date()

def yesterday():
    return prev_busday(today())


PAT_YEAR = re.compile("^\d{4}$")
PAT_MONTH = re.compile("^\d{4}M(0[1-9]{1}|10|11|12)$")
PAT_QUARTER = re.compile("^\d{4}Q[0-4]$")
PAT_HALF = re.compile("^\d{4}H[1-2]$")
PAT_WEEK = re.compile("^\d{4}W(0[1-9]|[1-4][0-9]|5[0-3]$)")
PAT_DAY = re.compile("^(D)(\d{4}-\d{1,2}-\d{1,2})")

#combine above 
PAT_PERIOD_ID = re.compile("|".join([PAT_YEAR.pattern, PAT_MONTH.pattern, PAT_QUARTER.pattern, PAT_HALF.pattern, PAT_WEEK.pattern, PAT_DAY.pattern]))


def is_period_id(text):
    if not text:
        return False

    match = PAT_PERIOD_ID.match(str(text))
    return (match is not None)


def start_of_period(period_id):
    pid = period_id
    assert is_period_id(pid)

    # daily
    if pid[0] == 'D':
        return prev_busday(parse(pid[1:]).date())

    year = int(pid[0:4])
    if len(pid) > 5: 
        part2 = int(pid[5:])  #month, week, quarter
        sep = pid[4]
    else:
        part2 = None
        sep = 'Y'
    if sep == 'W':        
        return isoweek.Week(year, part2).monday()
    elif sep == 'M':
        return datetime.date(year, part2, 1)
    elif sep == 'Q':
        month_id = (part2 - 1) * 3 + 1
        return datetime.date(year, month_id, 1)
    elif sep == 'H':
        return start_of_half(part2, year)
    elif sep == 'Y':
        return datetime.date(year, 1, 1)
    else:
        raise ValueError("Unexpected date identifier %s" % id)
    
def end_of_period(period_id):
    pid = period_id
    assert is_period_id(period_id)

    # daily
    if pid[0] == 'D':
        return parse(pid[1:]).date()


    year = int(period_id[0:4])
    if len(pid) > 5: 
        part2 = int(pid[5:])  #month, week, quarter
        sep = pid[4]
    else:
        part2 = None
        sep = 'Y'
    if sep == 'W':        
        return isoweek.Week(year, part2).sunday()
    elif sep == 'M':
        return end_of_month(part2, year)
    elif sep == 'Q':
        return end_of_quarter(part2, year)
    elif sep == 'H':
        return end_of_half(part2, year)
    elif sep == 'Y':
        return datetime.date(year, 12, 31)
    else:
        raise ValueError("Unexpected date identifier %s" % id)

def end_of_prev_period(period_id):
    return start_of_period(period_id) - datetime.timedelta(days=1)

##### END DOCENGINE  #####


from pandas.tseries.offsets import BDay


def prev_busday(d):
    if type(d) in [str, str]:
        d = parse(d).date()
    return (d - BDay(1)).date()


def as_date(thing):
    if isinstance(thing, datetime.date):
        return thing
    elif type(thing) == type(''):
        if '-' in thing:
            y,m,d = thing.split('-')
            return datetime.date(int(y), int(m), int(d))
    raise ValueError("Cannot represent %s as date" % repr(thing))


def start_of_next_month(d):
    if not isinstance(d, datetime.date):
        raise ValueError('Unexpected date: %s' % repr(d))
    year = d.year
    month = d.month + 1
    if month == 13:
        month = 1
        year += 1
    return datetime.date(year, month, 1)


def start_of_year(yr):
    return datetime.date(int(yr), 1, 1)

def start_of_prev_year(yr):
    return datetime.date(int(yr - 1), 1, 1)

def end_of_year(yr):
    return datetime.date(int(yr), 12, 31)

def end_of_prev_year(yr):
    return datetime.date(int(yr) - 1, 12, 31)

def end_of_month(mth,yr):
    mth = int(mth)
    yr = int(yr)
    if mth == 12:
        return datetime.date(int(yr) + 1, 1, 1) - datetime.timedelta(days=1)
    else:
        return datetime.date(int(yr), int(mth) + 1, 1) - datetime.timedelta(days=1)

def end_of_months(yr):
    return [end_of_month(mth, yr) for mth in range(1,13)]


def end_of_prev_month(mth,yr):
    return datetime.date(int(yr), int(mth), 1) - datetime.timedelta(days=1)

def start_of_month(mth,yr):
    return datetime.date(int(yr), int(mth), 1)


def MTD(dt):
    return start_of_month(dt.month, dt.year), dt

def QTD(dt):
    mm = dt.month
    if mm <=3:
        qtr = 1
    elif mm <=6:
        qtr = 2
    elif mm <=9:
        qtr = 3
    else:
        qtr = 4
    
    return start_of_quarter(qtr, dt.year), dt

def HTD(dt):
    mm = dt.month
    if mm <=6:
        half = 1
    else:
        half = 2
    
    return start_of_half(half), dt


def start_of_quarter(qtr, yr):
    return datetime.date(int(yr), int(qtr) * 3 - 2, 1)

def end_of_quarter(qtr, yr):
    return end_of_month(int(qtr) * 3, int(yr))

def end_of_prev_quarter(qtr, yr):
    start_of_qtr = start_of_quarter(qtr, yr)
    return start_of_qtr + datetime.timedelta(days=-1)

def prev_quarter(qtr, year):
    qtr = int(qtr)
    year = int(year)
    if qtr == 1:
        return 4, year-1
    else:
        return qtr-1, year


def prev_half(half, year):
    half = int(half)
    year = int(year)
    if half == 1:
        return 2, year-1
    else:
        return 1, year

def start_of_half(half, yr):
    return datetime.date(int(yr), int(half) * 6 - 5, 1)

def end_of_half(half, yr):
    return end_of_month(int(half) * 6, int(yr))

def end_of_prev_half(half, yr):
    return start_of_half(half, yr) + datetime.timedelta(days=-1)


def month_ends(yr):
    return [end_of_month(m, yr) for m in range(1,13)]

def monthrange(start, finish):
    months = (finish.year - start.year) * 12 + finish.month + 1
    for i in range(start.month, months):
        year  = (i - 1) / 12 + start.year
        month = (i - 1) % 12 + 1
        yield (year, month)

def quarterrange(start, end):
    start_qtr = (start.year, (start.month -1)/3 + 1)
    end_qtr = (end.year, (end.month -1)/3 + 1)
    quarters = (end_qtr[0] - start_qtr[0]) * 4 + (end_qtr[1] - start_qtr[1]) + 1
    for i in range(start_qtr[1], start_qtr[1] + quarters):
        year  = (i - 1) / 4 + start.year
        qtr = (i - 1) % 4 + 1
        yield (year, qtr)

def halfrange(start, end):
    start_half = (start.year, (start.month -1)/6 + 1)
    end_half = (end.year, (end.month -1)/6 + 1)
    semis = (end_half[0] - start_half[0]) * 2 + (end_half[1] - start_half[1]) + 1
    for i in range(start_half[1], start_half[1] + semis):
        year  = (i - 1) / 2 + start.year
        half = (i - 1) % 2 + 1
        yield (year, half)


def annualrange(start, end):
    years = end.year - start.year + 1
    for year in range(start.year, start.year + years):
        yield year



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('snapshot', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='glsnapshot',
            name='snapped_at',
            field=models.DateTimeField(help_text=b'Local NY time'),
        ),
    ]



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Log',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('level', models.CharField(max_length=10)),
                ('time', models.DateTimeField(auto_now_add=True)),
                ('message', models.TextField()),
                ('traceback', models.TextField()),
                ('request', models.TextField()),
            ],
            options={
                'ordering': ['-time', 'id'],
                'verbose_name_plural': 'Log entries',
            },
        ),
        migrations.CreateModel(
            name='TaskMonitor',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('task_id', models.CharField(max_length=200, null=True)),
                ('creation_timestamp', models.DateTimeField(auto_now_add=True, null=True)),
                ('task_name', models.CharField(max_length=200, null=True)),
                ('percent_complete', models.IntegerField(default=0, null=True)),
                ('task_state', models.CharField(max_length=200, null=True, choices=[(b'started', b'started'), (b'failed', b'failed'), (b'succeeded', b'succeeded')])),
                ('traceback', models.TextField(null=True)),
                ('log_info', models.TextField(null=True)),
                ('initiator', models.CharField(max_length=200, null=True)),
            ],
            options={
                'ordering': ['-creation_timestamp'],
            },
        ),
    ]



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reporting', '0005_reportdef_description'),
    ]

    operations = [
        migrations.AddField(
            model_name='reportdef',
            name='calc_type',
            field=models.CharField(default='as_of', max_length=20, choices=[(b'as_of', b'Point in Time'), (b'diff', b'Change Over Time')]),
            preserve_default=False,
        ),
    ]



import operator

from accountifie.common.api import api_func

def get_default_company():
    try:
        return api_func('environment', 'variable', 'DEFAULT_COMPANY_ID')
    except:
        return None

def get_trading_company():
    return api_func('environment', 'variable', 'TRADING_COMPANY_ID')
    
def get_alias(name):
    try:
        return api_func('environment', 'alias', name)['display_as']
    except:
        return name

def get_company(request):
    return request.session.get('company_id', get_default_company())

def get_company_name(company_id):
    company_names = dict((entry['id'], entry['name']) for entry in api_func('gl', 'company'))
    try:
        return company_names[company_id]
    except:
        return 'Unknown Company'

def find_first(path, acct_list):
    return min([int(x) for x in acct_list if path in acct_list[x]])

def order_paths(path_list):
    ACCT_LIST = dict((a['id'], a['path']) for a in api_func('gl', 'account'))
    d = dict((p, find_first(p, ACCT_LIST)) for p in path_list)
    return [x[0] for x in sorted(list(d.items()), key=operator.itemgetter(1))]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-11-27 01:10
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reporting', '0006_reportdef_calc_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='reportdef',
            name='calc_type',
            field=models.CharField(choices=[('as_of', 'Point in Time'), ('diff', 'Change Over Time')], max_length=20),
        ),
    ]



from decimal import Decimal
import csv

from django.contrib.auth.decorators import login_required
from django.conf import settings
from django.shortcuts import render
from django.template import RequestContext, Context
from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponse

from accountifie.cal.models import Year
from accountifie.common.api import api_func
import accountifie.reporting.apiv1 as rptg_api
import accountifie.query.query_manager
import accountifie.toolkit.utils as utils
import accountifie.reporting.rptutils as rptutils

import logging

logger = logging.getLogger('default')



ACCT_DEF = {
    'id': {'link': lambda x: "/api/reporting/transaction/%s" % x},
    'date': {},
    'comment': {},
    'account_id': {'link': lambda x: "/reporting/history/account/%s" % x},
    'contra_accts': {'link': lambda x: "/reporting/history/account/%s" % x},
    'counterparty': {'link': lambda x: "/reporting/history/creditor/%s" % x},
    'amount': {'fmt': lambda x: "{:,.2f}".format(Decimal(x))},
    'balance': {'fmt': lambda x: "{:,.2f}".format(Decimal(x))}
}

def create_entry(value, fmtr, qs):
    entry = {'text': fmtr['fmt'](value) if 'fmt' in fmtr else value}
    if 'link' in fmtr:
        entry['link'] = fmtr['link'](value)
        if qs:
            entry['link'] += '?%s' % qs
    return entry


def create_row(row, col_order, qs=None, fmtr=ACCT_DEF):
    return [create_entry(row[i], fmtr[i], qs=qs) for i in col_order]

@login_required
def history(request, type, id):
    qs = request.GET.urlencode()
    config = rptutils.history_prep(request)
    as_csv = (request.GET.get('as_csv', 'false').lower() == 'true')

    if type == 'account':
        acct = api_func('gl', 'account', id)
        display_name = '%s: %s' %(acct['id'], acct['display_name'])
        column_titles = ['id', 'date', 'comment', 'contra_accts', 'counterparty', 'amount', 'balance']
    elif type == 'path':
        id = id.replace('_', '.')
        display_name = 'path: %s' % id
        column_titles = ['id', 'date', 'comment', 'account_id', 'contra_accts', 'counterparty', 'amount', 'balance']
    elif type == 'creditor':
        cp_info = api_func('gl', 'counterparty', id)
        display_name = '%s: %s' %(cp_info['id'], cp_info['name'])
        column_titles = ['id', 'date', 'comment', 'contra_accts', 'amount', 'balance']
    else:
        raise ValueError('This type of history is not supported')
    

    hist_rslts = rptg_api.history(id, config)
    if as_csv:
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename="report.csv"'

        writer = csv.writer(response)
        writer.writerow(column_titles)
        for h in hist_rslts:
            writer.writerow([h[col] for col in column_titles])
        return response

    else:
        if hist_rslts:
            entries = [create_row(h, column_titles, qs=qs) for h in hist_rslts]
        else:
            entries = []
        
        context = {}
        context['display_name'] = display_name
        context['column_titles'] = column_titles
        context['history'] = entries
        context['years'] = Year.objects.all()
        context['by_date_cleared'] = False
        context['from_date'] = config.get('from', settings.DATE_EARLY)
        context['to_date'] = config.get('to', settings.DATE_LATE)
        return render(request, 'history.html', context)



import re
import datetime

from django.conf import settings
from dateutil.parser import parse
import accountifie.toolkit.utils.datefuncs as datefuncs


MONTHS = ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec']

MONTH_TAG = re.compile("^(\d{4})M(0[1-9]{1}|10|11|12)(Monthly)?$")
QUARTER_TAG = re.compile("^(\d{4})Q([1-4]{1})(Quarterly|Monthly)?$")
HALF_TAG = re.compile("^(\d{4})H([1-2]{1})(Semi|Quarterly|Monthly)?$")
YEAR_TAG = re.compile("^(\d{4})(Annual|Semi|Quarterly|Monthly)?$")


YTD_TAG = re.compile('(YTD_)(\d{4}-\d{1,2}-\d{1,2})(Annual|Semi|Quarterly|Monthly)?')
DAILY_TAG = re.compile('(daily_)(\d{4}-\d{1,2}-\d{1,2})')
DAY_TAG = re.compile('(day_)(\d{4}-\d{1,2}-\d{1,2})')
TRAILING12M_TAG = re.compile('(12Mtrailing)(-?)(\d{4}-\d{1,2}-\d{1,2})?')
MULTIYEAR_TAG = re.compile('(\d{1,2})yr_(\d{4}-\d{1,2}-\d{1,2})(Annual|Semi|Quarterly|Monthly)?')


BY_MAP = {'Monthly': 'month',
          'Quarterly': 'quarter',
          'Semi': 'half',
          'Annual': 'year'}


def date_from_shortcut(scut):
    if not scut:
        return None
    
    if scut == 'yesterday':
        return datefuncs.yesterday()
    elif scut == 'today':
        return datefuncs.today()
    elif scut == 'end_of_last_month':
        today = datefuncs.today()
        return datefuncs.end_of_prev_month(today.month, today.year)
    elif scut is None:
        return datefuncs.today()
    else:
        try:
            if type(scut) == datetime.date:
                return scut
            elif type(scut) in [str, str]:
                return parse(scut).date()
            else:
                return
        except:
            return


def parse_shortcut(col_tag):

    day_tag = DAY_TAG.match(col_tag)
    if day_tag is not None:
        dt = parse(day_tag.groups()[1]).date()
        return {'config_type': 'period',
                'period': 'day', 
                'year': dt.year,
                'month': dt.month,
                'day': dt.day,
                'by': 'day'}

    month_tag = MONTH_TAG.match(col_tag)
    if month_tag is not None:
        year = month_tag.groups()[0]
        month = month_tag.groups()[1]
        return {'config_type': 'period',
                'period': 'month', 
                'year': year,
                'month': month,
                'by': BY_MAP.get(month_tag.groups()[2], 'month')}

    quarter_tag = QUARTER_TAG.match(col_tag)
    if quarter_tag is not None:
        year = quarter_tag.groups()[0]
        quarter = quarter_tag.groups()[1]
        return {'config_type': 'period',
                'period': 'quarter', 
                'year': year,
                'quarter': quarter,
                'by': BY_MAP.get(quarter_tag.groups()[2], 'quarter')}

    half_tag = HALF_TAG.match(col_tag)
    if half_tag is not None:
        year = half_tag.groups()[0]
        half = half_tag.groups()[1]
        return {'config_type': 'period',
                'period': 'half', 
                'year': year,
                'half': half,
                'by': BY_MAP.get(half_tag.groups()[2], 'half')}

    year_tag = YEAR_TAG.match(col_tag)
    if year_tag is not None:
        year = year_tag.groups()[0]
        return {'config_type': 'period',
                'period': 'year', 
                'year': year,
                'by': BY_MAP.get(year_tag.groups()[1], 'year')}


    ytd_match = YTD_TAG.search(col_tag)
    if ytd_match:
        dt = parse(ytd_match.groups()[1]).date()
        return {'config_type': 'date_range',
                'from': datefuncs.end_of_prev_year(dt.year),
                'to': dt,
                'by': BY_MAP.get(ytd_match.groups()[2], 'year')}

    daily_match = DAILY_TAG.search(col_tag)
    if daily_match:
        dt = parse(daily_match.groups()[1]).date()
        return {'config_type': 'date_range',
                'from': datefuncs.prev_busday(dt),
                'to': dt,
                'by': 'day'}

    trailing12M_match = TRAILING12M_TAG.search(col_tag)
    if trailing12M_match:
        if trailing12M_match.groups()[2]:
            dt = date_from_shortcut(trailing12M_match.groups()[2])
        else:
            dt = datefuncs.today()
        return {'config_type': 'date_range',
                'from': datefuncs.end_of_month(dt.month, dt.year - 1),
                'to': dt,
                'by': 'month'}

    multiyear_match = MULTIYEAR_TAG.search(col_tag)
    if multiyear_match:
        dt = parse(multiyear_match.groups()[1]).date()
        years = int(multiyear_match.groups()[0])
        return {'config_type': 'date_range',
                'from': dt,
                'to': datefuncs.end_of_prev_month(dt.month, dt.year + years),
                'by': BY_MAP.get(multiyear_match.groups()[2], 'year')}

    if col_tag == 'current_YTD':
        return {'config_type': 'period',
                'period': 'year', 
                'year': datefuncs.today().year,
                'by': 'year'}

    if col_tag == 'current_HTD':
        start_of_half = datefuncs.HTD(datefuncs.today())[0]
        half_num = 1 if start_of_half.month == 1 else 2
        return {'config_type': 'period',
                'period': 'half', 
                'year': start_of_half.year,
                'half': half_num,
                'by': 'half'}

    if col_tag == 'current_QTD':
        start_of_qtr = datefuncs.QTD(datefuncs.today())[0]
        qtr_num = 1 + int(start_of_qtr.month - 1) / 3
        return {'config_type': 'period',
                'period': 'quarter', 
                'year': start_of_qtr.year,
                'quarter': qtr_num,
                'by': 'quarter'}

    if col_tag == 'current_MTD':
        today = datefuncs.today()
        return {'config_type': 'period',
                'period': 'month', 
                'year': today.year,
                'month': today.month,
                'by': 'month'}
    # didn't match anything
    raise ValueError('Unexpected shortcut: %s' % repr(col_tag))



from django.db import models
import datetime


class Metric(models.Model):
    name = models.CharField(max_length=50)

    def __str__(self):
        return self.name


class MetricEntry(models.Model):
    metric = models.ForeignKey(Metric, null=True, blank=True)
    label = models.CharField(max_length=20, null=True, blank=True)
    date = models.DateField(db_index=True)
    as_of = models.DateField(db_index=True)
    balance = models.DecimalField(max_digits=20, decimal_places=2)



from multipledispatch import dispatch

from django.forms import model_to_dict

from accountifie.snapshot.models import GLSnapshot
from accountifie.common.api import api_wrapper
import logging

logger = logging.getLogger('default')

def get_fld(record, fld):
    return getattr(record, fld)

def get_record_data(record, flds):
    data = dict((fld, get_fld(record, fld)) for fld in flds)
    return data



@dispatch(dict)
def glsnapshot(qstring):
    records = GLSnapshot.objects.order_by('id')
    flds = ['id', 'desc_link','short_desc', 'snapped_at', 'closing_date', 'comment']
    return [get_record_data(rec, flds) for rec in records]


@dispatch(str, dict)
def glsnapshot(snap_id, qstring):
    return model_to_dict(GLSnapshot.objects.get(id=snap_id))




from django.db import models
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist


class Address(models.Model):
    label = models.CharField(max_length=20)
    name = models.CharField(max_length=100, blank=True, null=True)
    company = models.CharField(max_length=100, blank=True, null=True)
    address1 = models.CharField(max_length=100, blank=True, null=True)
    address2 = models.CharField(max_length=100, blank=True, null=True)
    city = models.CharField(max_length=50, blank=True, null=True)
    postal_code = models.CharField(max_length=20, blank=True, null=True)
    province = models.CharField(max_length=30, blank=True, null=True)
    country = models.CharField(max_length=30, blank=True, null=True)
    phone = models.CharField(max_length=30, blank=True, null=True)
    email = models.EmailField(max_length=254, blank=True, null=True)

    def __unicode__(self):
        return self.label

    class Meta:
        app_label = 'common'
        db_table = 'common_address'
    

class McModel(models.Model):

    class Meta:
        abstract = True

    def to_json(self, expand=[]):
        flds = [f.name for f in self._meta.fields]
        properties = [p for p in getattr(self, 'properties', []) if p in expand]

        data = {}
        for fld in flds:
            if isinstance(self._meta.get_field_by_name(fld)[0], models.ForeignKey) and fld in expand:
                try:
                    data[fld] = getattr(self, fld).to_json(expand=expand)
                except:
                    data[fld] = str(getattr(self, fld))
            else:
                data[fld] = str(getattr(self, fld))

        for prop in properties:
            data[prop] = getattr(self, prop)
        return data


"""
Adapted with permission from ReportLab's DocEngine framework
"""


class Log(models.Model):
    '''Log data for any internal housekeeping'''
    level = models.CharField(max_length = 10)
    time = models.DateTimeField(auto_now_add = True, blank = True)
    message = models.TextField()
    traceback = models.TextField()
    request = models.TextField()

    class Meta:
        verbose_name_plural = 'Log entries'
        ordering = ['-time','id']
        app_label = 'common'
        db_table = 'common_log'

    def __unicode__(self):
            return 'Log("%s", %s, "%s...")' % (self.level, self.time, self.message[0:50])


class TaskMonitor(models.Model): 
    '''Keep track of task id's and the users associated with them'''
    TASK_STATES=(('started', 'started'),('failed', 'failed'), ('succeeded', 'succeeded'))
    task_id = models.CharField(max_length=200, null=True)
    creation_timestamp = models.DateTimeField(auto_now_add=True, null=True)
    task_name = models.CharField(max_length=200, null=True)
    percent_complete = models.IntegerField(null=True, default=0)
    task_state = models.CharField(max_length=200, null=True, choices=TASK_STATES)
    traceback = models.TextField(null=True)
    log_info = models.TextField(null=True)
    initiator = models.CharField(max_length=200, null=True)
    
    class Meta:
        ordering = ['-creation_timestamp']
        app_label = 'common'
        db_table = 'common_taskmonitor'

    def __unicode__(self):
        return "monitor for %s, %s" % (self.task_name, self.task_id)



"Re-saves all things which might produce GL transactions."
import inspect
import importlib
import logging
import datetime

from django.db import transaction
from django.http import HttpResponseRedirect

#from accountifie.gl.models import Transaction
from accountifie.gl.bmo import BusinessModelObject
from accountifie.common.api import api_func
from accountifie.celery import background_task
from accountifie.query.query_manager_strategy_factory import QueryManagerStrategyFactory

logger = logging.getLogger('default')


def recalculate(request):
    task_name = 'ledger-recalculate'
    task_id = background_task(task=task_name, calc=_recalculate).id
    status_url = 'background_task/status/%s' % task_id
    return HttpResponseRedirect('/tasks/list/')


def _recalculate(*args, **kwargs):
    klasses = []
    kl_paths = api_func('environment', 'variable_list', 'BMO_MODULES')

    # find all the BMO classes
    for path in kl_paths:
        for name, kl in inspect.getmembers(importlib.import_module(path), inspect.isclass):
            if BusinessModelObject in kl.__bases__:
                klasses.append(kl)

    with transaction.atomic():
        Transaction.objects.all().delete()

        for cmpny in [c['id'] for c in api_func('gl', 'company')]:
            QueryManagerStrategyFactory().erase(cmpny)

        logger.info("Recalculate -- deleted all transactions")
        QueryManagerStrategyFactory().set_fast_inserts('*', True)
        for klass in klasses:
            logger.info('Recalculate -- working on %s' % klass)
            qs =klass.objects.all()
            for obj in qs:
                try:
                    obj.update_gl()
                except:
                    logger.error('Recalc failed on %s - %s' % (klass._meta.model_name, str(obj)))
            logger.info('Recalculate -- finished %s' % klass)
        QueryManagerStrategyFactory().set_fast_inserts('*', False)
        QueryManagerStrategyFactory().take_snapshot('*')

    logger.info("Recalculate -- updated %d transactions" % qs.count())
    return



"Re-saves all things which might produce GL transactions."
import logging
from dateutil.parser import parse
import datetime

from django.http import HttpResponseRedirect
from django.contrib.auth.decorators import user_passes_test

from accountifie.common.models import Log
from accountifie.celery import background_task
from accountifie.query.query_manager import QueryManager
from accountifie.gl.models import Company
from . import utils


logger = logging.getLogger('default')


def cleanlogs(request):
    try:
        clean_to = parse(request.GET['clean_to'])
    except:
        clean_to = datetime.datetime.now() + datetime.timedelta(days=1)
    
    task_name = 'cleanlogs'
    task_id = background_task(task=task_name, calc=_cleanlogs, clean_to=clean_to).id
    return HttpResponseRedirect('/tasks/list')


def _cleanlogs(*args, **kwargs):
    clean_to = kwargs['clean_to']
    to_clean = Log.objects.filter(time__lte=clean_to)
    Log.objects.filter(time__lte=clean_to).delete()
    logger.info('cleaned %d log entries to %s' %(len(to_clean), clean_to.isoformat()))
    return


@user_passes_test(lambda u: u.is_superuser)
def primecache(request):
    task_name = 'primecache'
    task_id = background_task(task=task_name, calc=_run_primecache)
    logger.info(request, 'Balances Cache primed')
    return HttpResponseRedirect('/tasks/list')  


def _run_primecache(*args, **kwargs):
    year = datetime.datetime.now().year
    dates_list = utils.end_of_months(year) + utils.end_of_months(year-1)
    dates = dict((dt.isoformat(), dt) for dt in dates_list)
    company_ids = [c['id'] for c in Company.objects.filter(cmpy_type='ALO').values('id')]
    tags = [None, ['yearend']]
    query_manager = QueryManager()
    for cmpy in company_ids:
        for tag in tags:
            exclude_cps = [None]
            exclude_cps.append([c for c in company_ids if c != cmpy])
            for exclude_cp in exclude_cps:
                logger.info('priming balance cache with %s, %s, %s' %(str(cmpy), str(tag), str(exclude_cp)))
                throw_away = query_manager.pd_acct_balances(cmpy, dates, excl_contra=exclude_cp, excl_tags=tag)
    return





from django.shortcuts import render
from django.contrib.auth.decorators import login_required

from accountifie.common.table import get_table

@login_required
def tasks_list(request):
    context = {}
    context['title'] = 'Tasks'
    context['content'] = get_table('tasks_list')()
    return render(request, 'tasks_list.html', context)


from multipledispatch import dispatch

from django.db.models import Prefetch

from accountifie.reporting.models import *
from accountifie.reporting.serializers import *


def metric_labels(qstring):
    labels = MetricEntry.objects.all().values('label').distinct()
    return [x['label'] for x in labels if x['label'] not in [None, '']]


@dispatch(dict)
def metric(qstring):
    label = qstring.get('label', None)
    if type(label) == list:
        label = label[0]
    if label:
        qs = MetricEntry.objects.filter(label=label)
        return MetricEntrySerializer(qs, many=True).data
    else:
        return None


@dispatch(str, dict)
def metric(name, qstring):
    label = qstring.get('label', None)
    if type(label) == list:
        label = label[0]
    if label:
        qs = MetricEntry.objects.filter(metric__name=name) \
                                .filter(label=label)

        return MetricEntrySerializer(qs, many=True).data
    else:
        return None



from django.template import Context
from django.template.loader import get_template


def get_bstrap_table(data_url, row_defs, pagination="true", pagination_num=25):
    context = {}

    if '?' in data_url:
        context['data_url'] = data_url + '&raw=true'
    else:
        context['data_url'] = data_url + '?raw=true'

    context['row_defs'] = row_defs
    context['pagination'] = pagination
    context['pagination_num'] = pagination_num

    tmpl = get_template('bstrap_table.html')
    return tmpl.render(Context(context))


def get_modal(content, modal_title, modal_id):
    context = {}
    context['content'] = content
    context['modal_id'] = modal_id
    context['modal_title'] = modal_title

    tmpl = get_template('modal_cmpnt.html')
    return tmpl.render(Context(context))



import csv
import logging
import datetime

from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.shortcuts import render
from django.template import RequestContext

import accountifie.toolkit.utils as utils
from accountifie.common.api import api_func
from .models import Counterparty, Account

from accountifie.query.query_manager import QueryManager
from accountifie.query.query_manager_strategy_factory import QueryManagerStrategyFactory


fmt = lambda x: "{:,.0f}".format(x)
logger = logging.getLogger('default')


@login_required
def index(request):
    d = {}
    return render(request, 'index.html', d)


@login_required
def download_transactions(request):
    company_id = utils.get_company(request)

    #snapshot_time = datetime.datetime.now()
    strategy = QueryManagerStrategyFactory().get('remote')
    #strategy.set_cache(None)

    trans = strategy.get_all_transactions(company_id)

    all_accts_list = api_func('gl', 'account')
    all_accts = dict((r['id'], r) for r in all_accts_list)

    response = HttpResponse(content_type='text/csv')
    response['Content-Disposition'] = 'attachment; filename="transactions.csv"'
    writer = csv.writer(response)

    writer.writerow(['', '', '', '', '', '','debit', 'debit', 'debit', 'debit','credit', 'credit', 'credit', 'credit'])
    writer.writerow(['Id', 'dateEnd', 'date', 'type', 'comment', 'counterpartyId', 'amount', 'accountId',
                    'account name','counterpartyId', 'amount', 'accountId', 'account name'])

    for ex in trans:
        first_line = ex['lines'][0]
        acct_id = first_line['accountId']
        acct = api_func('gl', 'account', acct_id)
        if (acct['role'] in ['asset', 'expense'] and float(first_line['amount']) >0) or \
            (acct['role'] in ['liability', 'income', 'capital'] and float(first_line['amount']) < 0):
            debit = ex['lines'][0]
            credit = ex['lines'][1]
        else:
            debit = ex['lines'][1]
            credit = ex['lines'][0]

        row = [ex[k] for k in ['bmoId', 'dateEnd', 'date', 'type', 'comment']]
        row += [debit[k] for k in ['counterpartyId', 'amount', 'accountId']]
        row.append(all_accts[debit['accountId']]['display_name'])
        row += [credit[k] for k in ['counterpartyId', 'amount', 'accountId']]
        row.append(all_accts[credit['accountId']]['display_name'])

        writer.writerow(row)

    return response

@login_required
def download_tranlines(request):
    company_id = utils.get_company(request)
    trans = QueryManager().tranlines(company_id)

    response = HttpResponse(content_type='text/csv')
    response['Content-Disposition'] = 'attachment; filename="tranlines.csv"'
    writer = csv.writer(response)

    header_row = ['TranLine ID', 'Transaction ID', 'Amount', 'Account Name', 'Counterparty Name', 'Counterparty ID']

    writer.writerow(header_row)
    for ex in trans:
        writer.writerow(ex)

    return response



@login_required
def accounts_list(request):
    "Show list of each account"
    accounts = api_func('gl', 'account')
    return render(request, 'gl/accounts_list.html', dict(accounts=accounts))


@login_required
def counterparty_list(request):
    "Show list of each account"
    counterparties = api_func('gl', 'counterparty')
    AP_acct = api_func('environment', 'variable', 'GL_ACCOUNTS_PAYABLE')

    return render(request, 'gl/counterparty_list.html', dict(ap_acct=AP_acct, counterparties=counterparties))



"""
from django import forms
from dal import autocomplete

from .models import Counterparty, Account


def counterparty_widget():
    return forms.ModelChoiceField(queryset=Counterparty.objects.all(),
      widget=autocomplete.ModelSelect2(url='counterparty-autocomplete')
                                  )

def account_widget():
    return forms.ModelChoiceField(queryset=Account.objects.all(),
                                  widget=autocomplete.ModelSelect2(url='account-autocomplete')
                                  )
"""



import datetime
import mock
from django.test import TestCase

from accountifie.gl.models import Account, Company, Counterparty
from accountifie.reporting.core_reports.activity import AccountActivity
from accountifie.query.query_manager_strategy import QueryManagerStrategy

class ActivityTestCase(TestCase):
  def setUp(self):
    self.default_company = Company(name='TEST_COMPANY', id='TEST')
    self.default_company.save()

    Account(id='1001', path='test.one', display_name='TestOne', role='asset').save()
    Account(id='1002', path='test.two', display_name='TestTwo', role='expense').save()
    Account(id='1003', path='test.three', display_name='TestThree', role='liability').save()

  def test_activity_happy_path(self):
    # setup mock strategy
    mock_strategy = mock.create_autospec(QueryManagerStrategy, instance=True)
    mock_strategy.account_balances_for_dates.return_value = {
      '2014-12-31': {
        '1001': 10.00,
        '1002': 0.00,
        '1003': 24.00
      },
       'chg in period': {
        '1001': 100.00,
        '1002': -1.00,
        '1003': 20.00
       },
       '2015-12-31': {
        '1001': 110.00,
        '1002': -1.00,
        '1003': 44.00
      }
    }

    # create activity
    activity = AccountActivity('TEST')
    activity.configure({'col_tag': '2015Annual'})
    activity.set_gl_strategy(mock_strategy)

    # run calc
    result = activity.calcs()

    # test calls
    mock_strategy.account_balances_for_dates.assert_called_once_with(
      company_id='TEST',
      account_ids=['1001', '1002', '1003'],
      # TODO: remove hardcoded start date
      dates={'2014-12-31': {'start': datetime.date(2013, 1, 1), 'end': datetime.date(2014,12,31)},
            'chg in period': {'start': datetime.date(2015, 1, 1), 'end': datetime.date(2015,12,31)},
            '2015-12-31': {'start': datetime.date(2013, 1, 1), 'end': datetime.date(2015,12,31)},
      },
      with_counterparties=None,
      excl_interco=False,
      excl_contra=None,
      with_tags=None,
      excl_tags=None
    )

    # test result
    self.assertEqual(result[0]['label'], '1001: TestOne')
    self.assertEqual(result[0]['2014-12-31']['text'], 10.0)
    self.assertEqual(result[0]['chg in period']['text'], 100.0)
    self.assertEqual(result[0]['2015-12-31']['text'], 110.0)

    self.assertEqual(result[1]['label'], '1002: TestTwo')
    self.assertEqual(result[1]['2014-12-31']['text'], 0.0)
    self.assertEqual(result[1]['chg in period']['text'], -1.0)
    self.assertEqual(result[1]['2015-12-31']['text'], -1.0)

    self.assertEqual(result[2]['label'], '1003: TestThree')
    self.assertEqual(result[2]['2014-12-31']['text'], 24.0)
    self.assertEqual(result[2]['chg in period']['text'], 20.0)
    self.assertEqual(result[2]['2015-12-31']['text'], 44.0)
